# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for the LDAP user directory."""

import logging
import os
import re
import shutil
import subprocess

import augeas

from plinth import action_utils, utils
from plinth.actions import privileged

INPUT_LINES = None
ACCESS_CONF = '/etc/security/access.conf'
LDAPSCRIPTS_CONF = '/etc/ldapscripts/freedombox-ldapscripts.conf'


def _validate_user(username, password, must_be_admin=True):
    """Validate a user."""
    if must_be_admin:
        admins = _get_admin_users()

        if not admins:
            # any user is valid
            return

        if not username:
            raise PermissionError('Authentication user is required')

        if username not in admins:
            msg = f'"{username}" is not authorized to perform this action'
            raise PermissionError(msg)

    if not username:
        raise PermissionError('Authentication user is required')

    _validate_password(username, password)


def _validate_password(username, password):
    """Raise an error if the user password is invalid."""
    if not utils.is_authenticated_user(username, password):
        raise PermissionError('Invalid credentials')


@privileged
def first_setup():
    """Perform initial setup of LDAP."""
    # Avoid reconfiguration of slapd during module upgrades, because
    # this will move the existing database.
    # XXX: Instead of a separate action that is conditionally called for a
    # version number, we can check if the domain currently configured is what
    # we want and then based on the value do a reconfiguration. This approach
    # will work better when FreedomBox state is reset etc.
    action_utils.dpkg_reconfigure('slapd', {'domain': 'thisbox'})


@privileged
def setup():
    """Setup LDAP."""
    # Update pam configs for access and mkhomedir.
    subprocess.run(['pam-auth-update', '--package'], check=True)

    _configure_ldapscripts()

    _configure_ldap_authentication()

    _configure_ldap_structure()


def _configure_ldap_authentication():
    """Configure LDAP authentication."""
    action_utils.dpkg_reconfigure(
        'nslcd', {
            'ldap-uris': 'ldapi:///',
            'ldap-base': 'dc=thisbox',
            'ldap-auth-type': 'SASL',
            'ldap-sasl-mech': 'EXTERNAL'
        })
    action_utils.dpkg_reconfigure('libnss-ldapd',
                                  {'nsswitch': 'group, passwd, shadow'})
    action_utils.service_restart('nscd')

    # XXX: Workaround for login issue
    action_utils.service_enable('slapd')
    action_utils.service_start('slapd')
    action_utils.service_enable('nslcd')
    action_utils.service_start('nslcd')


def _configure_ldap_structure():
    """Configure LDAP basic structure."""
    was_running = action_utils.service_is_running('slapd')
    if not was_running:
        action_utils.service_start('slapd')

    _setup_admin()
    _create_organizational_unit('users')
    _create_organizational_unit('groups')


def _create_organizational_unit(unit):
    """Create an organizational unit in LDAP."""
    distinguished_name = 'ou={unit},dc=thisbox'.format(unit=unit)
    try:
        subprocess.run([
            'ldapsearch', '-Q', '-Y', 'EXTERNAL', '-H', 'ldapi:///', '-s',
            'base', '-b', distinguished_name, '(objectclass=*)'
        ], stdout=subprocess.DEVNULL, check=True)
        return  # Already exists
    except subprocess.CalledProcessError:
        input = '''
dn: ou={unit},dc=thisbox
objectClass: top
objectClass: organizationalUnit
ou: {unit}'''.format(unit=unit)
        subprocess.run(['ldapadd', '-Q', '-Y', 'EXTERNAL', '-H', 'ldapi:///'],
                       input=input.encode(), stdout=subprocess.DEVNULL,
                       check=True)


def _setup_admin():
    """Remove LDAP admin password and Allow root to modify the users."""
    process = subprocess.run([
        'ldapsearch', '-Q', '-L', '-L', '-L', '-Y', 'EXTERNAL', '-H',
        'ldapi:///', '-s', 'base', '-b', 'olcDatabase={1}mdb,cn=config',
        '(objectclass=*)', 'olcRootDN', 'olcRootPW'
    ], check=True, stdout=subprocess.PIPE)
    ldap_object = {}
    for line in process.stdout.decode().splitlines():
        if line:
            line = line.split(':')
            ldap_object[line[0]] = line[1]

    if 'olcRootPW' in ldap_object:
        subprocess.run(
            ['ldapmodify', '-Q', '-Y', 'EXTERNAL', '-H', 'ldapi:///'],
            check=True, stdout=subprocess.DEVNULL, input=b'''
dn: olcDatabase={1}mdb,cn=config
changetype: modify
delete: olcRootPW''')

    root_dn = 'gidNumber=0+uidNumber=0,cn=peercred,cn=external,cn=auth'
    if ldap_object['olcRootDN'] != root_dn:
        subprocess.run(
            ['ldapmodify', '-Q', '-Y', 'EXTERNAL', '-H', 'ldapi:///'],
            check=True, stdout=subprocess.DEVNULL, input=b'''
dn: olcDatabase={1}mdb,cn=config
changetype: modify
replace: olcRootDN
olcRootDN: gidNumber=0+uidNumber=0,cn=peercred,cn=external,cn=auth
''')


def _configure_ldapscripts():
    """Set the configuration used by ldapscripts for later user management."""
    # modify a copy of the config file
    shutil.copy('/etc/ldapscripts/ldapscripts.conf', LDAPSCRIPTS_CONF)

    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Shellvars/lens', 'Shellvars.lns')
    aug.set('/augeas/load/Shellvars/incl[last() + 1]', LDAPSCRIPTS_CONF)
    aug.load()

    # XXX: Password setting on users is disabled as changing passwords
    # using SASL Auth is not supported.
    aug.set('/files' + LDAPSCRIPTS_CONF + '/SERVER', '"ldapi://"')
    aug.set('/files' + LDAPSCRIPTS_CONF + '/SASLAUTH', '"EXTERNAL"')
    aug.set('/files' + LDAPSCRIPTS_CONF + '/SUFFIX', '"dc=thisbox"')
    aug.set('/files' + LDAPSCRIPTS_CONF + '/USUFFIX', '"ou=Users"')
    aug.set('/files' + LDAPSCRIPTS_CONF + '/GSUFFIX', '"ou=Groups"')
    aug.set('/files' + LDAPSCRIPTS_CONF + '/PASSWORDGEN', '"true"')
    aug.set('/files' + LDAPSCRIPTS_CONF + '/CREATEHOMES', '"yes"')
    aug.save()


@privileged
def get_nslcd_config() -> dict[str, str]:
    """Get nslcd configuration for diagnostics."""
    nslcd_conf = '/etc/nslcd.conf'
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.transform('Nslcd', nslcd_conf)
    aug.set('/augeas/context', '/files' + nslcd_conf)
    aug.load()

    return {
        'uri': aug.get('uri/1'),
        'base': aug.get('base'),
        'sasl_mech': aug.get('sasl_mech')
    }


def _get_samba_users():
    """Get users from the Samba user database."""
    # 'pdbedit -L' is better for listing users but is installed only with samba
    stdout = subprocess.check_output(
        ['tdbdump', '/var/lib/samba/private/passdb.tdb']).decode()
    return re.findall(r'USER_(.*)\\0', stdout)


def _delete_samba_user(username):
    """Delete a Samba user."""
    if username in _get_samba_users():
        subprocess.check_call(['smbpasswd', '-x', username])
        _disconnect_samba_user(username)


def _disconnect_samba_user(username):
    """Disconnect a Samba user."""
    try:
        subprocess.check_call(['pkill', '-U', username, 'smbd'])
    except subprocess.CalledProcessError as error:
        if error.returncode != 1:
            raise


@privileged
def create_user(username: str, password: str, auth_user: str | None = None,
                auth_password: str | None = None):
    """Create an LDAP user, set password and flush cache."""
    _validate_user(auth_user, auth_password)

    _run(['ldapadduser', username, 'users'])
    _set_user_password(username, password)
    _flush_cache()
    _set_samba_user(username, password)


@privileged
def remove_user(username: str, password: str | None = None):
    """Remove an LDAP user."""
    groups = _get_user_groups(username)

    # require authentication if the user is last admin user
    if _get_group_users('admin') == [username]:
        _validate_password(username, password)

    _delete_samba_user(username)

    for group in groups:
        _remove_user_from_group(username, group)

    _run(['ldapdeleteuser', username])

    _flush_cache()


@privileged
def rename_user(old_username: str, new_username: str):
    """Rename an LDAP user."""
    groups = _get_user_groups(old_username)

    _delete_samba_user(old_username)

    for group in groups:
        _remove_user_from_group(old_username, group)

    _run(['ldaprenameuser', old_username, new_username])

    for group in groups:
        _add_user_to_group(new_username, group)

    _flush_cache()


def _set_user_password(username, password):
    """Set a user's password."""
    process = _run(['slappasswd', '-s', password], stdout=subprocess.PIPE)
    password = process.stdout.decode().strip()
    _run(['ldapsetpasswd', username, password])


def _set_samba_user(username, password):
    """Insert a user to the Samba database.

    If a user already exists, update password.
    """
    proc = subprocess.run(['smbpasswd', '-a', '-s', username],
                          input='{0}\n{0}\n'.format(password).encode(),
                          stderr=subprocess.PIPE, check=False)
    if proc.returncode != 0:
        raise RuntimeError('Unable to add Samba user: ', proc.stderr)


@privileged
def set_user_password(username: str, password: str, auth_user: str,
                      auth_password: str):
    """Set a user's password."""
    must_be_admin = username != auth_user
    _validate_user(auth_user, auth_password, must_be_admin=must_be_admin)

    _set_user_password(username, password)
    _set_samba_user(username, password)


def _get_admin_users():
    """Return list of members in the admin group.

    Raise an error if the slapd service is not running.
    """
    admin_users = []

    try:
        output = subprocess.check_output([
            'ldapsearch', '-LLL', '-Q', '-Y', 'EXTERNAL', '-H', 'ldapi:///',
            '-o', 'ldif-wrap=no', '-s', 'base', '-b',
            'cn=admin,ou=groups,dc=thisbox', 'memberUid'
        ]).decode()
    except subprocess.CalledProcessError as error:
        if error.returncode == 32:
            # no entries found
            return []
        raise

    for line in output.splitlines():
        if line.startswith('memberUid: '):
            user = line.split('memberUid: ', 1)[1].strip()
            admin_users.append(user)

    return admin_users


def _get_group_users(groupname):
    """Return list of members in the group."""
    try:
        process = _run(['ldapgid', '-P', groupname], stdout=subprocess.PIPE)
    except subprocess.CalledProcessError:
        return []  # Group does not exist

    output = process.stdout.decode()
    # extract users from output, example: 'admin:*:10001:user1,user2'
    users = output.rsplit(':')[-1].strip().split(',')
    if users == ['']:
        return []
    return users


def _get_user_groups(username):
    """Return only the supplementary groups of the given user.

    Exclude the 'users' primary group from the returned list.
    """
    process = _run(['ldapid', username], stdout=subprocess.PIPE, check=False)
    output = process.stdout.decode().strip()
    if output:
        groups_part = output.split(' ')[2]
        try:
            groups = groups_part.split('=')[1]
        except IndexError:
            logging.warning('Could not read groups for user %s: \n%s',
                            username, output)
            return []

        group_names = [
            user.strip('()') for user in re.findall(r'\(.*?\)', groups)
        ]
        group_names.remove('users')
        return group_names

    logging.warning('User %s not found in LDAP', username)
    return []


@privileged
def get_user_groups(username: str) -> list[str]:
    """Return list of a given user's groups."""
    return _get_user_groups(username)


def _group_exists(groupname):
    """Return whether a group already exits."""
    process = _run(['ldapgid', groupname], check=False)
    return process.returncode == 0


def _create_group(groupname):
    """Add an LDAP group."""
    if not _group_exists(groupname):
        _run(['ldapaddgroup', groupname])


@privileged
def create_group(groupname: str):
    """Add an LDAP group."""
    _create_group(groupname)
    _flush_cache()


@privileged
def rename_group(old_groupname: str, new_groupname: str):
    """Rename an LDAP group.

    Skip if the group to rename from doesn't exist.
    """
    if old_groupname == 'admin' or new_groupname == 'admin':
        raise ValueError('Can\'t rename the group "admin"')

    if _group_exists(old_groupname):
        _run(['ldaprenamegroup', old_groupname, new_groupname])
        _flush_cache()


@privileged
def remove_group(groupname: str):
    """Remove an LDAP group."""
    if groupname == 'admin':
        raise ValueError("Can't remove the group 'admin'")

    if _group_exists(groupname):
        _run(['ldapdeletegroup', groupname])
        _flush_cache()


def _add_user_to_group(username, groupname):
    """Add an LDAP user to an LDAP group."""
    _create_group(groupname)
    _run(['ldapaddusertogroup', username, groupname])


@privileged
def add_user_to_group(username: str, groupname: str,
                      auth_user: str | None = None,
                      auth_password: str | None = None):
    """Add an LDAP user to an LDAP group."""
    if groupname == 'admin':
        _validate_user(auth_user, auth_password)

    _add_user_to_group(username, groupname)
    _flush_cache()


def _remove_user_from_group(username, groupname):
    """Remove an LDAP user from an LDAP group."""
    _run(['ldapdeleteuserfromgroup', username, groupname])


@privileged
def remove_user_from_group(username: str, groupname: str, auth_user: str,
                           auth_password: str):
    """Remove an LDAP user from an LDAP group."""
    if groupname == 'admin':
        _validate_user(auth_user, auth_password)

    _remove_user_from_group(username, groupname)
    _flush_cache()
    if groupname == 'freedombox-share':
        _disconnect_samba_user(username)


@privileged
def get_group_users(group_name: str) -> list[str]:
    """Get the list of users of an LDAP group."""
    return _get_group_users(group_name)


@privileged
def set_user_status(username: str, status: str, auth_user: str,
                    auth_password: str):
    """Set the status of the user."""
    if status not in ('active', 'inactive'):
        raise ValueError('Invalid status')

    _validate_user(auth_user, auth_password)

    if status == 'active':
        flag = '-e'
    else:
        flag = '-d'

    if username in _get_samba_users():
        subprocess.check_call(['smbpasswd', flag, username])
        if status == 'inactive':
            _disconnect_samba_user(username)


def _flush_cache():
    """Flush nscd and apache2 cache."""
    _run(['nscd', '--invalidate=passwd'])
    _run(['nscd', '--invalidate=group'])
    action_utils.service_reload('apache2')


def _run(arguments, check=True, **kwargs):
    """Run a command. Check return code and suppress output by default."""
    env = dict(os.environ, LDAPSCRIPTS_CONF=LDAPSCRIPTS_CONF)
    kwargs['stdout'] = kwargs.get('stdout', subprocess.DEVNULL)
    kwargs['stderr'] = kwargs.get('stderr', subprocess.DEVNULL)
    return subprocess.run(arguments, env=env, check=check, **kwargs)



# SPDX-License-Identifier: AGPL-3.0-or-later

backup = {
    'config': {
        'directories': ['/etc/ez-ipupdate/']
    },
    'settings': [
        'dynamicdns_enable', 'dynamicdns_config', 'dynamicdns_status'
    ],
}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for privileged snapshot operations.
"""

import pathlib

import pytest

from plinth.modules.snapshot import privileged

systemctl_path = pathlib.Path('/usr/bin/systemctl')
systemd_installed = pytest.mark.skipif(not systemctl_path.exists(),
                                       reason='systemd not available')


@pytest.mark.parametrize('input_path,escaped_path',
                         [('.snapshots', '\\x2esnapshots'), ('/', '-'),
                          ('/home/user', 'home-user'), (':_.', ':_.')])
@systemd_installed
def test_systemd_path_escape(input_path, escaped_path):
    """Test escaping systemd paths."""
    assert escaped_path == privileged._systemd_path_escape(input_path)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Views for the email app.
"""

from django.contrib import messages
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views.generic.base import TemplateView
from django.views.generic.edit import FormView

from plinth.modules.email import dns
from plinth.views import AppView

from . import aliases as aliases_module
from . import forms, privileged


class EmailAppView(AppView):
    """Server configuration page."""
    app_id = 'email'
    form_class = forms.DomainForm
    template_name = 'email.html'

    def get_context_data(self, **kwargs):
        """Add additional context data for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['dns_entries'] = dns.get_entries()
        return context

    def get_initial(self):
        """Return the initial values to populate in the form."""
        initial = super().get_initial()
        domains = privileged.domain.get_domains()
        initial['primary_domain'] = domains['primary_domain']
        return initial

    def form_valid(self, form):
        """Update the settings for changed domain values."""
        old_data = form.initial
        new_data = form.cleaned_data
        if old_data['primary_domain'] != new_data['primary_domain']:
            try:
                privileged.domain.set_all_domains(new_data['primary_domain'])
                messages.success(self.request, _('Configuration updated'))
            except Exception:
                messages.error(self.request,
                               _('An error occurred during configuration.'))

        return super().form_valid(form)


class AliasView(FormView):
    """View to create, list, enable, disable and delete aliases.

    This view has two forms. Form to list (and manage) existing aliases, and a
    form to create a new aliases. When GET operation is used, both forms
    created and template is rendered. When POST operation is used, the form
    posted is detected using hidden form values and the appropriate form is
    initialized for the FormView base class to work with.
    """
    template_name = 'email-aliases.html'
    form_classes = (forms.AliasCreateForm, forms.AliasListForm)
    success_url = reverse_lazy('email:aliases')

    def __init__(self, *args, **kwargs):
        """Initialize the view."""
        super().__init__(*args, **kwargs)
        self.posted_form = None

    def get_form_class(self):
        """Return form class to build."""
        if self.posted_form == 'create':
            return forms.AliasCreateForm

        return forms.AliasListForm

    def get_form_kwargs(self):
        """Send aliases to list form."""
        kwargs = super().get_form_kwargs()
        if self.posted_form != 'create':
            kwargs['aliases'] = self._get_current_aliases()

        return kwargs

    def _get_current_aliases(self):
        """Return current list of aliases."""
        return aliases_module.get(self.request.user.username)

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        if isinstance(context['form'], forms.AliasCreateForm):
            context['create_form'] = context['form']
            aliases = self._get_current_aliases()
            context['list_form'] = forms.AliasListForm(aliases)
        else:
            context['create_form'] = forms.AliasCreateForm()
            context['list_form'] = context['form']

        return context

    def post(self, request, *args, **kwargs):
        """Find which form was submitted before proceeding."""
        self.posted_form = request.POST.get('form')
        return super().post(request, *args, **kwargs)

    def form_valid(self, form):
        """Handle a valid submission."""
        if isinstance(form, forms.AliasListForm):
            self._list_form_valid(form)
        elif isinstance(form, forms.AliasCreateForm):
            self._create_form_valid(form)

        return super().form_valid(form)

    def _list_form_valid(self, form):
        """Handle a valid alias list form operation."""
        aliases = form.cleaned_data['aliases']
        username = self.request.user.username
        aliases_module.delete(username, aliases)

    def _create_form_valid(self, form):
        """Handle a valid create alias form operation."""
        username = self.request.user.username
        aliases_module.put(username, form.cleaned_data['alias'])


class XmlView(TemplateView):
    """Serve Thunderbird auto-configuration file.

    https://wiki.mozilla.org/Thunderbird:Autoconfiguration:ConfigFileFormat
    """
    template_name = 'email-autoconfig.xml'

    def render_to_response(self, *args, **kwargs):
        kwargs['content_type'] = 'text/xml; charset=utf-8'
        return super().render_to_response(*args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['host'] = self.request.get_host()
        return context



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the radicale module.
"""

from django.urls import re_path

from .views import RadicaleAppView

urlpatterns = [
    re_path(r'^apps/radicale/$', RadicaleAppView.as_view(), name='index'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""App component for other apps to use Apache configuration functionality."""

import re
import subprocess

from django.utils.translation import gettext_noop

from plinth import action_utils, app
from plinth.diagnostic_check import (DiagnosticCheck,
                                     DiagnosticCheckParameters, Result)
from plinth.privileged import service as service_privileged

from . import privileged


class Webserver(app.LeaderComponent):
    """Component to enable/disable Apache configuration."""

    def __init__(self, component_id, web_name, kind='config', urls=None,
                 expect_redirects=False, last_updated_version=None):
        """Initialize the web server component.

        component_id should be a unique ID across all components of an app and
        across all components.

        web_name is the primary part of the configuration file path which must
        be enabled/disabled by this component.

        kind is the type of Apache configuration being enabled/disabled. This
        must be 'config' for a configuration in /etc/apache/conf-available/,
        'module' for configuration in /etc/apache2/mods-available/, 'site' for
        configuration in /etc/apache2/sites-available/.

        urls is a list of URLs over which a HTTP services will be available due
        to this component. This list is only used for running diagnostics.

        last_updated_version is the app version in which the web server
        configuration/site/module file was updated. Using this, web server will
        be automatically reloaded or restarted as necessary during app upgrade.
        """
        super().__init__(component_id)

        self.web_name = web_name
        self.kind = kind
        self.urls = urls or []
        self.expect_redirects = expect_redirects
        self.last_updated_version = last_updated_version or 0

    def is_enabled(self):
        """Return whether the Apache configuration is enabled."""
        return action_utils.webserver_is_enabled(self.web_name, kind=self.kind)

    def enable(self):
        """Enable the Apache configuration."""
        privileged.enable(self.web_name, self.kind)

    def disable(self):
        """Disable the Apache configuration."""
        privileged.disable(self.web_name, self.kind)

    def diagnose(self) -> list[DiagnosticCheck]:
        """Check if the web path is accessible by clients.

        See :py:meth:`plinth.app.Component.diagnose`.

        """
        results = []
        for url in self.urls:
            if '{host}' in url:
                results.extend(
                    diagnose_url_on_all(url, check_certificate=False,
                                        expect_redirects=self.expect_redirects,
                                        component_id=self.component_id))
            else:
                results.append(
                    diagnose_url(url, check_certificate=False,
                                 component_id=self.component_id))

        return results

    def setup(self, old_version):
        """Restart/reload web server if configuration files changed."""
        if not old_version:
            # App is being freshly setup. After setup, app will be enabled
            # which will result in reload/restart of web server.
            return

        if old_version >= self.last_updated_version:
            # Already using the latest configuration. Web server reload/restart
            # is not necessary.
            return

        if not self.app.is_enabled():
            # App is currently disabled, web server will reloaded/restarted
            # when the app is enabled.
            return

        if self.kind == 'module':
            service_privileged.restart('apache2')
        else:
            service_privileged.reload('apache2')


class Uwsgi(app.LeaderComponent):
    """Component to enable/disable uWSGI configuration."""

    def __init__(self, component_id, uwsgi_name):
        """Initialize the uWSGI component.

        component_id should be a unique ID across all components of an app and
        across all components.

        uwsgi_name is the primary part of the configuration file path which
        must be enabled/disabled by this component.

        """
        super().__init__(component_id)

        self.uwsgi_name = uwsgi_name

    def is_enabled(self):
        """Return whether the uWSGI configuration is enabled."""
        return action_utils.uwsgi_is_enabled(self.uwsgi_name) \
            and action_utils.service_is_enabled('uwsgi')

    def enable(self):
        """Enable the uWSGI configuration."""
        privileged.uwsgi_enable(self.uwsgi_name)

    def disable(self):
        """Disable the uWSGI configuration."""
        privileged.uwsgi_disable(self.uwsgi_name)

    def is_running(self):
        """Return whether the uWSGI daemon is running with configuration."""
        return action_utils.uwsgi_is_enabled(self.uwsgi_name) \
            and action_utils.service_is_running('uwsgi')


def diagnose_url(url: str, kind: str | None = None,
                 env: dict[str, str] | None = None,
                 check_certificate: bool = True,
                 extra_options: list[str] | None = None,
                 wrapper: str | None = None,
                 expected_output: str | None = None,
                 component_id: str | None = None) -> DiagnosticCheck:
    """Run a diagnostic on whether a URL is accessible.

    Kind can be '4' for IPv4 or '6' for IPv6.
    """
    try:
        return_value = check_url(url, kind, env, check_certificate,
                                 extra_options, wrapper, expected_output)
        result = Result.PASSED if return_value else Result.FAILED
    except FileNotFoundError:
        result = Result.ERROR

    parameters: DiagnosticCheckParameters = {'url': url, 'kind': kind}
    if kind:
        check_id = f'apache-url-kind-{url}-{kind}'
        description = gettext_noop('Access URL {url} on tcp{kind}')
    else:
        check_id = f'apache-url-{url}'
        description = gettext_noop('Access URL {url}')

    return DiagnosticCheck(check_id, description, result, parameters,
                           component_id)


def diagnose_url_on_all(url: str, expect_redirects: bool = False,
                        component_id: str | None = None,
                        **kwargs) -> list[DiagnosticCheck]:
    """Run a diagnostic on whether a URL is accessible."""
    results = []
    for address in action_utils.get_addresses():
        current_url = url.format(host=address['url_address'])
        diagnose_kwargs = dict(kwargs)
        if not expect_redirects:
            diagnose_kwargs.setdefault('kind', address['kind'])

        results.append(
            diagnose_url(current_url, component_id=component_id,
                         **diagnose_kwargs))

    return results


def check_url(url: str, kind: str | None = None,
              env: dict[str, str] | None = None,
              check_certificate: bool = True,
              extra_options: list[str] | None = None,
              wrapper: str | None = None,
              expected_output: str | None = None) -> bool:
    """Check whether a URL is accessible."""
    command = ['curl', '--location', '-f', '-w', '%{response_code}']

    if kind == '6':
        # extract zone index
        match = re.match(r'(.*://)\[(.*)%(?P<zone>.*)\](.*)', url)
        if match:
            command = command + ['--interface', match.group('zone')]
            url = '{0}[{1}]{2}'.format(*match.group(1, 2, 4))

    command.append(url)

    if wrapper:
        command.insert(0, wrapper)

    if not check_certificate:
        command.append('-k')

    if extra_options:
        command.extend(extra_options)

    if kind:
        command.append({'4': '-4', '6': '-6'}[kind])

    try:
        process = subprocess.run(command, env=env, check=True,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
        result = True
        if expected_output and expected_output not in process.stdout.decode():
            result = False
    except subprocess.CalledProcessError as exception:
        result = False
        # Authorization failed is a success
        if exception.stdout.decode().strip() in ('401', '405'):
            result = True

    return result



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for Janus app.
"""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.janus]


class TestJanusApp(functional.BaseAppTests):
    app_name = 'janus'
    has_service = True
    has_web = True



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Views for the dynamicsdns module.
"""

import datetime

from django.contrib import messages
from django.utils.translation import gettext_lazy as _

from plinth import views
from plinth.modules import dynamicdns

from .forms import ConfigureForm


class DynamicDNSAppView(views.AppView):
    """Serve configuration page."""
    app_id = 'dynamicdns'
    template_name = 'dynamicdns.html'
    form_class = ConfigureForm

    _error_messages = {
        'timeout': _('Connection timed out'),
        'gaierror': _('Could not find server'),
        'TimeoutError': _('Connection timed out'),
        'ConnectionRefusedError': _('Server refused connection'),
        'ValueError': _('Already up-to-date')
    }

    def get_context_data(self, **kwargs):
        """Return the context data for rendering the template view."""
        context = super().get_context_data(**kwargs)
        status = dynamicdns.get_status()
        config = dynamicdns.get_config()
        domains_status = {}
        for domain_name, domain in status['domains'].items():
            if domain_name not in config['domains']:
                continue

            # Create naive datetime object in local timezone
            domain['timestamp'] = datetime.datetime.fromtimestamp(
                domain['timestamp'])
            domains_status[domain_name] = domain
            if domain['error_code'] in self._error_messages:
                domain['error_message'] = self._error_messages[
                    domain['error_code']]

        context['domains_status'] = domains_status
        return context

    def get_initial(self):
        """Get the current values for the form."""
        initial = super().get_initial()
        domains = dynamicdns.get_config()['domains']
        domain = list(domains.values())[0] if domains else {}
        initial.update(domain)
        return domain

    def form_valid(self, form):
        """Apply the changes submitted in the form."""
        old_status = form.initial
        new_status = form.cleaned_data

        if old_status != new_status:
            config = dynamicdns.get_config()
            try:
                del config['domains'][old_status['domain']]
            except KeyError:
                pass

            config['domains'][new_status['domain']] = new_status
            dynamicdns.set_config(config)
            if old_status.get('domain'):
                dynamicdns.notify_domain_removed(old_status['domain'])

            dynamicdns.notify_domain_added(new_status['domain'])
            messages.success(self.request, _('Configuration updated'))

        # Perform an immediate update, even when configuration is not changed.
        dynamicdns.update_dns(None)

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure disks manager."""

import os
import re
import stat
import subprocess

from plinth import utils
from plinth.actions import privileged


@privileged
def is_partition_expandable(device: str) -> int:
    """Return a list of partitions that can be expanded."""
    _, _, free_space = _get_free_space(device)
    return int(free_space['size'])


@privileged
def expand_partition(device: str, mount_point: str = '/'):
    """Expand a partition to take adjacent free space."""
    device, requested_partition, free_space = _get_free_space(device)

    if requested_partition['table_type'] == 'msdos' and \
       int(requested_partition['number']) >= 5:
        raise RuntimeError(
            'Expanding logical partitions currently unsupported')

    if requested_partition['table_type'] == 'gpt':
        _move_gpt_second_header(device)

    _resize_partition(device, requested_partition, free_space)
    _resize_file_system(device, requested_partition, free_space, mount_point)


def _move_gpt_second_header(device):
    """Move second header to the end of the disk.

    GPT scheme has two mostly identical partition table headers. One at the
    beginning of the disk and one at the end. When an image is written to
    larger disk, the second header is not at the end of the disk. Fix that by
    moving second partition to end of the disk before attempting partition
    resize.

    """
    command = ['sgdisk', '--move-second-header', device]
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError:
        raise RuntimeError('Error moving GPT second header to the end')


def _resize_partition(device, requested_partition, free_space):
    """Resize the partition table entry."""
    command = [
        'parted', '--align=optimal', '--script', device, 'unit', 'B',
        'resizepart', requested_partition['number'],
        str(free_space['end'])
    ]
    # XXX: Remove workaround after bug in parted is fixed:
    # https://debbugs.gnu.org/cgi/bugreport.cgi?bug=24215
    fallback_command = [
        'parted', '--align=optimal', device, '---pretend-input-tty', 'unit',
        'B', 'resizepart', requested_partition['number']
    ]
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError:
        try:
            input_text = 'yes\n' + str(free_space['end'])
            subprocess.run(fallback_command, check=True,
                           input=input_text.encode())
        except subprocess.CalledProcessError as exception:
            raise RuntimeError(f'Error expanding partition: {exception}')


def _resize_file_system(device, requested_partition, free_space,
                        mount_point='/'):
    """Resize a file system inside a partition."""
    if requested_partition['type'] == 'btrfs':
        _resize_btrfs(device, requested_partition, free_space, mount_point)
    elif requested_partition['type'] == 'ext4':
        _resize_ext4(device, requested_partition, free_space, mount_point)


def _resize_ext4(device, requested_partition, _free_space, _mount_point):
    """Resize an ext4 file system inside a partition."""
    partition_device = _get_partition_device(device,
                                             requested_partition['number'])
    try:
        command = ['resize2fs', partition_device]
        subprocess.run(command, stdout=subprocess.DEVNULL,
                       stderr=subprocess.DEVNULL, check=True)
    except subprocess.CalledProcessError as exception:
        raise RuntimeError(f'Error expanding filesystem: {exception}')


def _resize_btrfs(_device, _requested_partition, _free_space, mount_point='/'):
    """Resize a btrfs file system inside a partition."""
    try:
        command = ['btrfs', 'filesystem', 'resize', 'max', mount_point]
        subprocess.run(command, stdout=subprocess.DEVNULL, check=True)
    except subprocess.CalledProcessError as exception:
        raise RuntimeError(f'Error expanding filesystem: {exception}')


def _get_free_space(device):
    """Return the amount of free space after a partition."""
    device, partition_number = \
        _get_root_device_and_partition_number(device)

    try:
        requested_partition, free_spaces = \
            _get_partitions_and_free_spaces(device, partition_number)
    except Exception as exception:
        raise RuntimeError(f'Error getting partition details: {exception}')

    # Don't accept extended partitions for now
    if requested_partition['table_type'] == 'msdos' and \
       int(requested_partition['number']) >= 5:
        raise RuntimeError(
            'Expanding logical partitions currently unsupported')

    # Don't accept anything but btrfs and ext4 filesystems
    if requested_partition['type'] not in ('btrfs', 'ext4'):
        raise RuntimeError(
            f'Unsupported file system type: {requested_partition["type"]}')

    found_free_space = None
    for free_space in free_spaces:
        if free_space['start'] != requested_partition['end'] + 1:
            continue

        if free_space['size'] < 10 * 1024 * 1024:  # Minimum 10MiB
            continue

        found_free_space = free_space

    if not found_free_space:
        raise RuntimeError('No free space available')

    return device, requested_partition, found_free_space


def _get_partition_device(device, partition_number):
    """Return the device corresponding to a parition in a given device."""
    if re.match('[0-9]', device[-1]):
        return device + 'p' + str(partition_number)

    return device + str(partition_number)


def _get_root_device_and_partition_number(device):
    """Return the parent device and number of partition separately."""
    match = re.match(r'(.+[a-zA-Z]\d+)p(\d+)$', device)
    if not match:
        match = re.match(r'(.+[a-zA-Z])(\d+)$', device)
        if not match:
            raise ValueError('Invalid device, must be a partition')

    return match.group(1), match.group(2)


def _get_partitions_and_free_spaces(device, partition_number):
    """Run parted and return list of partitions and free spaces."""
    command = [
        'parted', '--machine', '--script', device, 'unit', 'B', 'print', 'free'
    ]
    process = subprocess.run(command, stdout=subprocess.PIPE, check=True)

    requested_partition = None
    free_spaces = []

    lines = process.stdout.decode().splitlines()
    partition_table_type = lines[1].split(':')[5]
    for line in lines[2:]:
        line = line.rstrip(';')
        keys = ('number', 'start', 'end', 'size', 'type')
        parts = line.split(':')
        segment = dict(zip(keys, parts[:5]))

        segment['table_type'] = partition_table_type
        segment['start'] = _interpret_unit(segment['start'])
        segment['end'] = _interpret_unit(segment['end'])
        segment['size'] = _interpret_unit(segment['size'])

        if segment['type'] == 'free':
            segment['number'] = None
            free_spaces.append(segment)
        else:
            if segment['number'] == partition_number:
                requested_partition = segment

    return requested_partition, free_spaces


def _interpret_unit(value):
    """Return value in bytes after understanding parted unit."""
    value = value.rstrip('B')  # For now, we only need to understand bytes
    return int(value)


@privileged
def mount(block_device: str):
    """Mount a disk are root user.

    XXX: This is primarily to provide compatibility with older code that used
    udiskie to auto-mount all partitions as root user under /media/root/
    directory. We are setting special permissions for the directory /media/root
    and users have set shared folders using this path. This can be removed in
    favor of using DBus API once we have a migration plan in place. Disks can
    be mounted directly /mount without ACL restrictions that apply to
    /mount/<user> directories. This can be done by setting udev flag
    UDISKS_FILESYSTEM_SHARED=1 by writing a udev rule.

    """
    subprocess.run([
        'udisksctl', 'mount', '--block-device', block_device,
        '--no-user-interaction'
    ], check=True)


@privileged
def eject(device_path: str) -> str:
    """Eject a device by its path."""
    return _eject_drive_of_device(device_path)


def _get_options():
    """Return the common options used for udisks2 operations."""
    glib = utils.import_from_gi('GLib', '2.0')
    options = glib.Variant(
        'a{sv}', {'auth.no_user_interaction': glib.Variant('b', True)})
    return options


def _eject_drive_of_device(device_path):
    """Eject a device after unmounting all of its partitions.

    Return the details (model, vendor) of drives ejected.
    """
    udisks = utils.import_from_gi('UDisks', '2.0')
    glib = utils.import_from_gi('GLib', '2.0')
    client = udisks.Client.new_sync()
    object_manager = client.get_object_manager()

    found_objects = [
        obj for obj in object_manager.get_objects()
        if obj.get_block() and obj.get_block().props.device == device_path
    ]

    if not found_objects:
        raise ValueError(
            'No such device - {device_path}'.format(device_path=device_path))

    obj = found_objects[0]

    # Unmount filesystems
    block_device = obj.get_block()
    drive_object_path = block_device.props.drive
    if drive_object_path != '/':
        _umount_all_filesystems_of_drive(drive_object_path)
    else:
        # Block device has not associated drive
        _umount_filesystem(obj.get_filesystem())

    # Eject the drive
    drive = client.get_drive_for_block(block_device)
    if drive:
        try:
            drive.call_eject_sync(_get_options(), None)
        except glib.Error:
            # Ignore error during ejection as along as all the filesystems are
            # unmounted, the disk can be removed.
            pass

        return {
            'vendor': drive.props.vendor,
            'model': drive.props.model,
        }

    return None


def _umount_filesystem(filesystem):
    """Unmount a filesystem."""
    if filesystem and filesystem.props.mount_points:
        filesystem.call_unmount_sync(_get_options())


def _umount_all_filesystems_of_drive(drive_object_path):
    """Unmount all filesystems on block devices of a drive."""
    udisks = utils.import_from_gi('UDisks', '2.0')
    client = udisks.Client.new_sync()
    object_manager = client.get_object_manager()

    for obj in object_manager.get_objects():
        block_device = obj.get_block()
        if not block_device or block_device.props.drive != drive_object_path:
            continue

        _umount_filesystem(obj.get_filesystem())


@privileged
def setup():
    """Configure storage."""
    # create udisks2 default mount directory
    mounts_directory = '/media/root'
    try:
        os.mkdir(mounts_directory)
    except FileExistsError:
        pass

    # make the directory readable and traversible by other users
    stats = os.stat(mounts_directory)
    os.chmod(mounts_directory, stats.st_mode | stat.S_IROTH | stat.S_IXOTH)


@privileged
def usage_info() -> str:
    """Get information about disk space usage."""
    command = [
        'df', '--exclude-type=tmpfs', '--exclude-type=devtmpfs',
        '--block-size=1', '--output=source,fstype,size,used,avail,pcent,target'
    ]
    return subprocess.check_output(command).decode()


@privileged
def validate_directory(directory: str, check_creatable: bool,
                       check_writable: bool):
    """Validate a directory."""
    if os.geteuid() == 0:
        raise RuntimeError('You must not be root to run this command')

    def part_exists(path):
        """Return part of the path that exists."""
        if not path or os.path.exists(path):
            return path
        return part_exists(os.path.dirname(path))

    if check_creatable:
        directory = part_exists(directory)
        if not directory:
            directory = '.'
    else:
        if not os.path.exists(directory):
            raise FileNotFoundError

    if not os.path.isdir(directory):
        raise NotADirectoryError

    if not os.access(directory, os.R_OK):
        raise PermissionError('read')

    if check_writable or check_creatable:
        if not os.access(directory, os.W_OK):
            raise PermissionError('write')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Remove the deprecated KVStore entries 'setup_state' and 'firstboot_state',
and only use the new entry 'firstboot_completed' instead.
"""

from __future__ import unicode_literals

from django.db import migrations

from plinth.models import KVStore


def merge_firstboot_finished_fields(apps, schema_editor):
    """
    Merge 'setup_state' and 'firstboot_state' into 'firstboot_completed'.

    'firstboot_completed' is the most accurate name for now, and by combining
    the fields we do not have to deal with legacy states/fields anymore.
    """
    # Get and remove 'firstboot_state'
    firstboot_state = 0
    try:
        _object = KVStore.objects.get(key='firstboot_state')
    except KVStore.DoesNotExist:
        pass
    else:
        firstboot_state = _object.value
        _object.delete()

    # Get and remove 'setup_state'
    setup_state = 0
    try:
        _object = KVStore.objects.get(key='setup_state')
    except KVStore.DoesNotExist:
        pass
    else:
        setup_state = _object.value
        _object.delete()

    # Get current 'firstboot_completed'
    firstboot_completed = False
    try:
        _object = KVStore.objects.get(key='firstboot_completed')
    except KVStore.DoesNotExist:
        pass
    else:
        firstboot_completed = _object.value

    # Set new 'firstboot_completed' if needed
    new_firstboot_completed = bool(firstboot_completed or setup_state
                                   or firstboot_state)
    if new_firstboot_completed and not firstboot_completed:
        obj, created = KVStore.objects.get_or_create(key='firstboot_completed')
        obj.value = 1
        obj.save()


class Migration(migrations.Migration):

    dependencies = [
        ('plinth', '0002_modulestore'),
    ]

    operations = [
        migrations.RunPython(merge_firstboot_finished_fields),
    ]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to manage backup archives."""

import json
import logging
import os
import pathlib
import re

import paramiko
from django.utils.text import get_valid_filename
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop

from plinth import app as app_module
from plinth import cfg, glib, menu
from plinth.package import Packages

from . import api, privileged

logger = logging.getLogger(__name__)

_description = [
    _('Backups allows creating and managing backup archives.'),
]

# session variable name that stores when a backup file should be deleted
SESSION_PATH_VARIABLE = 'fbx-backups-upload-path'


class BackupsApp(app_module.App):
    """FreedomBox app for backup and restore."""

    app_id = 'backups'

    _version = 3

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(
            app_id=self.app_id, version=self._version, is_essential=True,
            depends=['storage'], name=_('Backups'), icon='fa-files-o',
            description=_description, manual_page='Backups',
            donation_url='https://www.borgbackup.org/support/fund.html')
        self.add(info)

        menu_item = menu.Menu('menu-backups', info.name, None, info.icon,
                              'backups:index', parent_url_name='system:data',
                              order=20)
        self.add(menu_item)

        packages = Packages('packages-backups', ['borgbackup', 'sshfs'])
        self.add(packages)

    @staticmethod
    def post_init():
        """Perform post initialization operations."""
        # Check every hour to perform scheduled backups
        glib.schedule(3600, backup_by_schedule)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        from . import repository
        privileged.setup(repository.RootBorgRepository.PATH)
        self.enable()

        # First time setup or upgrading from older versions.
        if old_version <= 2:
            _show_schedule_setup_notification()


def _backup_handler(packet, encryption_passphrase=None):
    """Perform backup operation on packet."""
    if not os.path.exists(privileged.MANIFESTS_FOLDER):
        os.makedirs(privileged.MANIFESTS_FOLDER)

    manifest_path = os.path.join(privileged.MANIFESTS_FOLDER,
                                 get_valid_filename(packet.path) + '.json')
    manifests = {
        'apps': [{
            'name': component.app.app_id,
            'version': component.app.info.version,
            'backup': component.manifest
        } for component in packet.components]
    }
    with open(manifest_path, 'w', encoding='utf-8') as manifest_file:
        json.dump(manifests, manifest_file)

    paths = packet.directories + packet.files
    paths.append(manifest_path)

    privileged.create_archive(packet.path, paths,
                              comment=packet.archive_comment,
                              encryption_passphrase=encryption_passphrase)


def backup_by_schedule(data):
    """Check if backups need to be taken and run the operation."""
    from . import repository as repository_module
    for repository in repository_module.get_repositories():
        try:
            repository.schedule.run_schedule()
            _show_schedule_error_notification(repository, is_error=False)
        except Exception as exception:
            logger.exception('Error running scheduled backup: %s', exception)
            _show_schedule_error_notification(repository, is_error=True,
                                              exception=exception)


def _restore_exported_archive_handler(packet, encryption_passphrase=None):
    """Perform restore operation on packet."""
    privileged.restore_exported_archive(packet.path, packet.directories,
                                        packet.files)


def restore_archive_handler(packet, encryption_passphrase=None):
    """Perform restore operation on packet."""
    privileged.restore_archive(packet.path, '/', packet.directories,
                               packet.files, encryption_passphrase)


def restore_from_upload(path, app_ids=None):
    """Restore files from an uploaded .tar.gz backup file"""
    api.restore_apps(_restore_exported_archive_handler, app_ids=app_ids,
                     create_subvolume=False, backup_file=path)


def get_known_hosts_path():
    """Return the path to the known hosts file."""
    return pathlib.Path(cfg.data_dir) / '.ssh' / 'known_hosts'


def is_ssh_hostkey_verified(hostname):
    """Check whether SSH Hostkey has already been verified.

    hostname: Domain name or IP address of the host

    """
    known_hosts_path = get_known_hosts_path()
    if not known_hosts_path.exists():
        return False

    known_hosts = paramiko.hostkeys.HostKeys(str(known_hosts_path))
    host_keys = known_hosts.lookup(hostname)
    return host_keys is not None


def split_path(path):
    """Splits the given path into username, hostname, directory.

    Network interface information is kept in the hostname if provided.
    e.g. fe80::2078:6c26:498a:1fa5%wlp1s0

    """
    return re.findall(r'^(.*)@([^/]*):(.*)$', path)[0]


def _show_schedule_setup_notification():
    """Show a notification hinting to setup a remote backup schedule."""
    from plinth.notification import Notification
    message = gettext_noop(
        'Enable an automatic backup schedule for data safety. Prefer an '
        'encrypted remote backup location or an extra attached disk.')
    data = {
        'app_name': 'translate:' + gettext_noop('Backups'),
        'app_icon': 'fa-files-o'
    }
    title = gettext_noop('Enable a Backup Schedule')
    actions_ = [{
        'type': 'link',
        'class': 'primary',
        'text': gettext_noop('Go to {app_name}'),
        'url': 'backups:index'
    }, {
        'type': 'dismiss'
    }]
    Notification.update_or_create(id='backups-remote-schedule',
                                  app_id='backups', severity='info',
                                  title=title, message=message,
                                  actions=actions_, data=data, group='admin')


def on_schedule_save(repository):
    """Dismiss notification. Called when repository's schedule is updated."""
    if not repository.schedule.enabled:
        return

    from plinth.notification import Notification
    try:
        note = Notification.get('backups-remote-schedule')
        note.dismiss()
    except KeyError:
        pass


def _show_schedule_error_notification(repository, is_error, exception=None):
    """Show or hide a notification related scheduled backup operation."""
    from plinth.notification import Notification
    id_ = 'backups-schedule-error-' + repository.uuid
    try:
        note = Notification.get(id_)
        error_count = note.data['error_count']
    except KeyError:
        error_count = 0

    message = gettext_noop(
        'A scheduled backup failed. Past {error_count} attempts for backup '
        'did not succeed. The latest error is: {error_message}')
    data = {
        'app_name': 'translate:' + gettext_noop('Backups'),
        'app_icon': 'fa-files-o',
        'error_count': error_count + 1 if is_error else 0,
        'error_message': str(exception)
    }
    title = gettext_noop('Error During Backup')
    actions_ = [{
        'type': 'link',
        'class': 'primary',
        'text': gettext_noop('Go to {app_name}'),
        'url': 'backups:index'
    }, {
        'type': 'dismiss'
    }]
    note = Notification.update_or_create(id=id_, app_id='backups',
                                         severity='error', title=title,
                                         message=message, actions=actions_,
                                         data=data, group='admin')
    note.dismiss(should_dismiss=not is_error)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test scheduling of backups.
"""

import json
from datetime import datetime, timedelta
from unittest.mock import MagicMock, call, patch

import pytest

import plinth.modules.backups.repository  # noqa, pylint: disable=unused-import
from plinth.app import App

from ..components import BackupRestore
from ..schedule import Schedule


class AppTest(App):
    """Sample App for testing."""
    app_id = 'test-app'


def _get_backup_component(name):
    """Return a BackupRestore component."""
    return BackupRestore(name)


def _get_test_app(name):
    """Return an App."""
    app = AppTest()
    app.app_id = name
    app._all_apps[name] = app
    app.add(BackupRestore(name + '-component'))
    return app


def test_init_default_values():
    """Test initialization of schedule with default values."""
    schedule = Schedule('test-uuid')
    assert schedule.repository_uuid == 'test-uuid'
    assert not schedule.enabled
    assert schedule.daily_to_keep == 5
    assert schedule.weekly_to_keep == 3
    assert schedule.monthly_to_keep == 3
    assert schedule.run_at_hour == 2
    assert schedule.unselected_apps == []


def test_init():
    """Test initialization with explicit values."""
    schedule = Schedule('test-uuid', enabled=True, daily_to_keep=1,
                        weekly_to_keep=2, monthly_to_keep=5, run_at_hour=0,
                        unselected_apps=['test-app1', 'test-app2'])
    assert schedule.repository_uuid == 'test-uuid'
    assert schedule.enabled
    assert schedule.daily_to_keep == 1
    assert schedule.weekly_to_keep == 2
    assert schedule.monthly_to_keep == 5
    assert schedule.run_at_hour == 0
    assert schedule.unselected_apps == ['test-app1', 'test-app2']


def test_get_storage_format():
    """Test that storage format is properly returned."""
    schedule = Schedule('test-uuid', enabled=True, daily_to_keep=1,
                        weekly_to_keep=2, monthly_to_keep=5, run_at_hour=23,
                        unselected_apps=['test-app1', 'test-app2'])
    assert schedule.get_storage_format() == {
        'enabled': True,
        'daily_to_keep': 1,
        'weekly_to_keep': 2,
        'monthly_to_keep': 5,
        'run_at_hour': 23,
        'unselected_apps': ['test-app1', 'test-app2'],
    }


def _get_archives_from_test_data(data):
    """Return a list of archives from test data."""
    archives = []
    for index, item in enumerate(data):
        archive_time = item['time']
        if isinstance(archive_time, str):
            archive_time = datetime.strptime(archive_time,
                                             '%Y-%m-%d %H:%M:%S+0000')
        archive = {
            'comment':
                json.dumps({
                    'type': 'scheduled',
                    'periods': item['periods']
                }),
            'start':
                archive_time.strftime('%Y-%m-%dT%H:%M:%S.%f'),
            'name':
                f'archive-{index}'
        }
        archives.append(archive)

    return archives


# - First item is the arguments to send construct Schedule()
# - Second item is the list of previous backups in the system.
# - Third item is the return value of datetime.datetime.now().
# - Fourth item is the list of periods for which backups must be triggered.
# - Fifth item is the list of expected archives to be deleted after backup.
cases = [
    # Schedule is disabled
    [
        [False, 10, 10, 10, 0],
        [],
        datetime(2021, 1, 1),
        [],
        [],
    ],
    # No past backups
    [
        [True, 10, 10, 10, 0],
        [],
        datetime(2021, 1, 1),
        ['daily', 'weekly', 'monthly'],
        [],
    ],
    # Daily backup taken recently
    [
        [True, 10, 10, 10, 0],
        [{
            'periods': ['daily'],
            'time': datetime(2021, 1, 1) - timedelta(seconds=600)
        }],
        datetime(2021, 1, 1),
        [],
        [],
    ],
    # Weekly backup taken recently
    [
        [True, 10, 10, 10, 0],
        [{
            'periods': ['weekly'],
            'time': datetime(2021, 1, 1) - timedelta(seconds=600)
        }],
        datetime(2021, 1, 1),
        [],
        [],
    ],
    # Monthly backup taken recently
    [
        [True, 10, 10, 10, 0],
        [{
            'periods': ['monthly'],
            'time': datetime(2021, 1, 1) - timedelta(seconds=600)
        }],
        datetime(2021, 1, 1),
        [],
        [],
    ],
    # Backup taken not so recently
    [
        [True, 10, 10, 10, 0],
        [{
            'periods': ['daily'],
            'time': datetime(2021, 1, 1) - timedelta(seconds=3 * 3600)
        }],
        datetime(2021, 1, 1),
        ['daily', 'weekly', 'monthly'],
        [],
    ],
    # Too long since a daily backup, not scheduled time
    [
        [True, 10, 10, 10, 2],
        [{
            'periods': ['daily'],
            'time': datetime(2021, 1, 1) - timedelta(days=1, seconds=3601)
        }],
        datetime(2021, 1, 1),
        ['daily', 'weekly', 'monthly'],
        [],
    ],
    # No too long since a daily backup, not scheduled time
    [
        [True, 10, 10, 10, 2],
        [{
            'periods': ['daily'],
            'time': datetime(2021, 1, 1) - timedelta(days=1, seconds=3600)
        }],
        datetime(2021, 1, 1),
        ['weekly', 'monthly'],
        [],
    ],
    # Too long since a weekly backup, not scheduled time
    [
        [True, 10, 10, 10, 2],
        [{
            'periods': ['weekly'],
            'time': datetime(2021, 1, 1) - timedelta(days=7, seconds=3601)
        }],
        datetime(2021, 1, 1),
        ['daily', 'weekly', 'monthly'],
        [],
    ],
    # No too long since a daily backup, not scheduled time
    [
        [True, 10, 10, 10, 2],
        [{
            'periods': ['weekly'],
            'time': datetime(2021, 1, 1) - timedelta(days=7, seconds=3600)
        }],
        datetime(2021, 1, 1),
        ['daily', 'monthly'],
        [],
    ],
    # Too long since a monthly backup, not scheduled time, year rounding
    [
        [True, 10, 10, 10, 2],
        [{
            'periods': ['monthly'],
            'time': datetime(2020, 12, 1)
        }],
        datetime(2021, 1, 1, 1, 0, 1),
        ['daily', 'weekly', 'monthly'],
        [],
    ],
    # No too long since a monthly backup, not scheduled time, year rounding
    [
        [True, 10, 10, 10, 2],
        [{
            'periods': ['monthly'],
            'time': datetime(2020, 12, 1)
        }],
        datetime(2021, 1, 1, 1),
        ['daily', 'weekly'],
        [],
    ],
    # Too long since a monthly backup, not scheduled time, no year rounding
    [
        [True, 10, 10, 10, 2],
        [{
            'periods': ['monthly'],
            'time': datetime(2020, 11, 1)
        }],
        datetime(2020, 12, 1, 1, 0, 1),
        ['daily', 'weekly', 'monthly'],
        [],
    ],
    # No too long since a monthly backup, not scheduled time, no year rounding
    [
        [True, 10, 10, 10, 2],
        [{
            'periods': ['monthly'],
            'time': datetime(2020, 11, 1)
        }],
        datetime(2020, 12, 1, 1),
        ['daily', 'weekly'],
        [],
    ],
    # Time for daily backup
    [
        [True, 10, 10, 10, 0],
        [{
            'periods': ['daily', 'weekly', 'monthly'],
            'time': datetime(2021, 1, 2) - timedelta(seconds=3 * 3600)
        }],
        datetime(2021, 1, 2),
        ['daily'],
        [],
    ],
    # Time for daily backup, different scheduled time
    [
        [True, 10, 10, 10, 11],
        [{
            'periods': ['daily', 'weekly', 'monthly'],
            'time': datetime(2021, 1, 2, 11) - timedelta(seconds=3 * 3600)
        }],
        datetime(2021, 1, 2, 11),
        ['daily'],
        [],
    ],
    # Time for daily/weekly backup, 2021-01-03 is a Sunday
    [
        [True, 10, 10, 10, 0],
        [{
            'periods': ['daily', 'weekly', 'monthly'],
            'time': datetime(2021, 1, 3) - timedelta(seconds=3 * 3600)
        }],
        datetime(2021, 1, 3),
        ['daily', 'weekly'],
        [],
    ],
    # Time for daily/monthly backup
    [
        [True, 10, 10, 10, 0],
        [{
            'periods': ['daily', 'weekly', 'monthly'],
            'time': datetime(2021, 1, 1) - timedelta(seconds=3 * 3600)
        }],
        datetime(2021, 1, 1),
        ['daily', 'monthly'],
        [],
    ],
    # Daily backups disabled by setting the no. of backups to keep to 0
    [
        [True, 0, 10, 10, 0],
        [],
        datetime(2021, 1, 1),
        ['weekly', 'monthly'],
        [],
    ],
    # Weekly backups disabled by setting the no. of backups to keep to 0
    [
        [True, 10, 0, 10, 0],
        [],
        datetime(2021, 1, 1),
        ['daily', 'monthly'],
        [],
    ],
    # Monthly backups disabled by setting the no. of backups to keep to 0
    [
        [True, 10, 10, 0, 0],
        [],
        datetime(2021, 1, 1),
        ['daily', 'weekly'],
        [],
    ],
    # Not scheduled, not too long since last, no backup necessary
    [
        [True, 10, 10, 10, 0],
        [{
            'periods': ['daily', 'weekly', 'monthly'],
            'time': datetime(2021, 1, 2, 1) - timedelta(seconds=3 * 3600)
        }],
        datetime(2021, 1, 2, 1),
        [],
        [],
    ],
    # Cleanup daily backups
    [
        [True, 2, 10, 10, 0],
        [{
            'periods': ['daily'],
            'time': datetime(2021, 1, 3, 1) - timedelta(seconds=3 * 3600)
        }, {
            'periods': ['daily'],
            'time': datetime(2021, 1, 2, 1) - timedelta(seconds=3 * 3600)
        }, {
            'periods': ['daily'],
            'time': datetime(2021, 1, 1, 1) - timedelta(seconds=3 * 3600)
        }, {
            'periods': ['daily'],
            'time': datetime(2021, 1, 1, 0) - timedelta(seconds=3 * 3600)
        }],
        datetime(2021, 1, 4, 1),
        ['daily', 'weekly', 'monthly'],
        ['archive-2', 'archive-3'],
    ],
    # Cleanup weekly backups
    [
        [True, 10, 2, 10, 0],
        [{
            'periods': ['weekly'],
            'time': datetime(2021, 1, 3, 1) - timedelta(seconds=3 * 3600)
        }, {
            'periods': ['weekly'],
            'time': datetime(2021, 1, 2, 1) - timedelta(seconds=3 * 3600)
        }, {
            'periods': ['weekly'],
            'time': datetime(2021, 1, 1, 1) - timedelta(seconds=3 * 3600)
        }, {
            'periods': ['weekly'],
            'time': datetime(2021, 1, 1, 0) - timedelta(seconds=3 * 3600)
        }],
        datetime(2021, 1, 4, 1),
        ['daily', 'monthly'],
        ['archive-2', 'archive-3'],
    ],
    # Cleanup monthly backups
    [
        [True, 10, 10, 2, 0],
        [{
            'periods': ['monthly'],
            'time': datetime(2021, 1, 3, 1) - timedelta(seconds=3 * 3600)
        }, {
            'periods': ['monthly'],
            'time': datetime(2021, 1, 2, 1) - timedelta(seconds=3 * 3600)
        }, {
            'periods': ['monthly'],
            'time': datetime(2021, 1, 1, 1) - timedelta(seconds=3 * 3600)
        }, {
            'periods': ['monthly'],
            'time': datetime(2021, 1, 1, 0) - timedelta(seconds=3 * 3600)
        }],
        datetime(2021, 1, 4, 1),
        ['daily', 'weekly'],
        ['archive-2', 'archive-3'],
    ],
    # Cleanup daily backups, but keep due to them being weekly/monthly too
    [
        [True, 2, 1, 10, 0],
        [{
            'periods': ['daily'],
            'time': datetime(2021, 1, 6, 1) - timedelta(seconds=3 * 3600)
        }, {
            'periods': ['daily'],
            'time': datetime(2021, 1, 5, 1) - timedelta(seconds=3 * 3600)
        }, {
            'periods': ['daily', 'weekly'],
            'time': datetime(2021, 1, 4, 1) - timedelta(seconds=3 * 3600)
        }, {
            'periods': ['daily', 'weekly'],
            'time': datetime(2021, 1, 3, 1) - timedelta(seconds=3 * 3600)
        }, {
            'periods': ['daily', 'monthly'],
            'time': datetime(2021, 1, 2, 0) - timedelta(seconds=3 * 3600)
        }, {
            'periods': ['daily'],
            'time': datetime(2021, 1, 1, 0) - timedelta(seconds=3 * 3600)
        }],
        datetime(2021, 1, 7, 1),
        ['daily'],
        ['archive-3', 'archive-5'],
    ],
]


@pytest.mark.parametrize(
    'schedule_params,archives_data,test_now,run_periods,cleanups', cases)
@patch('plinth.app.App.get_setup_state')
@patch('plinth.modules.backups.repository.get_instance')
def test_run_schedule(get_instance, get_setup_state, schedule_params,
                      archives_data, test_now, run_periods, cleanups):
    """Test that backups are run at expected time."""
    get_setup_state.return_value = App.SetupState.UP_TO_DATE

    repository = MagicMock()
    repository.list_archives.side_effect = \
        lambda: _get_archives_from_test_data(archives_data)
    get_instance.return_value = repository

    with patch('plinth.modules.backups.schedule.datetime') as mock_datetime, \
            patch('plinth.app.App.list') as app_list:
        app_list.return_value = [
            _get_test_app('test-app1'),
            _get_test_app('test-app2'),
            _get_test_app('test-app3')
        ]

        mock_datetime.now.return_value = test_now
        mock_datetime.strptime = datetime.strptime
        mock_datetime.min = datetime.min
        mock_datetime.side_effect = lambda *args, **kwargs: datetime(
            *args, **kwargs)

        schedule = Schedule('test_uuid', schedule_params[0],
                            schedule_params[1], schedule_params[2],
                            schedule_params[3], schedule_params[4],
                            ['test-app2'])
        schedule.run_schedule()

        if not run_periods:
            repository.create_archive.assert_not_called()
        else:
            run_periods.sort()
            name = 'scheduled: {periods}: {datetime}'.format(
                periods=', '.join(run_periods),
                datetime=mock_datetime.now().strftime('%Y-%m-%d:%H:%M'))
            app_ids = ['test-app1', 'test-app3']
            archive_comment = json.dumps({
                'type': 'scheduled',
                'periods': run_periods
            }).replace('{', '{{').replace('}', '}}')
            repository.create_archive.assert_has_calls(
                [call(name, app_ids, archive_comment=archive_comment)])

        if not cleanups:
            repository.delete_archive.assert_not_called()
        else:
            calls = [call(name) for name in cleanups]
            repository.delete_archive.assert_has_calls(calls)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for janus.
"""

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import frontpage, menu
from plinth.config import DropinConfigs
from plinth.daemon import Daemon
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.coturn.components import TurnTimeLimitedConsumer
from plinth.modules.firewall.components import Firewall
from plinth.package import Packages, install
from plinth.utils import Version, format_lazy

from . import manifest, privileged

_description = [
    _('Janus is a lightweight WebRTC server.'),
    _('A simple video conference room is included.'),
    format_lazy(
        _('<a href="{coturn_url}">Coturn</a> is required to '
          'use Janus.'), coturn_url=reverse_lazy('coturn:index')),
]


class JanusApp(app_module.App):
    """FreedomBox app for janus."""

    app_id = 'janus'

    _version = 2

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(self.app_id, self._version, name=_('Janus'),
                               icon_filename='janus',
                               short_description=_('Video Room'),
                               description=_description, manual_page='Janus',
                               clients=manifest.clients)
        self.add(info)

        menu_item = menu.Menu('menu-janus', info.name, info.short_description,
                              info.icon_filename, 'janus:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut('shortcut-janus', info.name,
                                      info.short_description,
                                      info.icon_filename,
                                      reverse_lazy('janus:room'),
                                      clients=manifest.clients)
        self.add(shortcut)

        packages = Packages('packages-janus', [
            'janus', 'libjs-bootbox', 'libjs-bootstrap', 'libjs-bootswatch',
            'libjs-janus-gateway', 'libjs-jquery-blockui', 'libjs-spin.js',
            'libjs-toastr', 'libjs-webrtc-adapter'
        ])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-janus', [
            '/etc/apache2/conf-available/janus-freedombox.conf',
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-janus', info.name,
                            ports=['http', 'https',
                                   'janus-freedombox'], is_external=True)
        self.add(firewall)

        webserver = Webserver('webserver-janus', 'janus-freedombox')
        self.add(webserver)

        daemon = Daemon(
            'daemon-janus', 'janus', listen_ports=[(8088, 'tcp4'),
                                                   (8088, 'tcp6'),
                                                   (8188, 'tcp4'),
                                                   (8188, 'tcp6')])
        self.add(daemon)

        turn = TurnTimeLimitedConsumer('turn-janus')
        self.add(turn)

        backup_restore = BackupRestore('backup-restore-janus',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if not old_version:
            self.enable()

    def force_upgrade(self, packages):
        """Force upgrade janus to resolve conffile prompts."""
        if 'janus' not in packages:
            return False

        # Allow upgrades within 1.0.* and 1.1.*
        package = packages['janus']
        if Version(package['new_version']) > Version('1.2~'):
            return False

        install(['janus'], force_configuration='new')
        privileged.setup()
        return True



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Forms for configuring ikiwiki
"""

from django import forms
from django.utils.translation import gettext_lazy as _


class IkiwikiCreateForm(forms.Form):
    """Form to create a wiki or blog."""
    site_type = forms.ChoiceField(label=_('Type'), choices=[('wiki', 'Wiki'),
                                                            ('blog', 'Blog')])

    name = forms.CharField(label=_('Name'))

    admin_name = forms.CharField(label=_('Admin Account Name'))

    admin_password = forms.CharField(label=_('Admin Account Password'),
                                     widget=forms.PasswordInput())



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Utility methods for managing translations.
"""

from django.conf import settings
from django.utils import translation


def _should_use_sessions(request):
    return hasattr(request, 'session') and hasattr(translation,
                                                   'LANGUAGE_SESSION_KEY')


def get_language_from_request(request):
    """Get the language in the session or as separate cookie.

    Django methods should be used for regular cases. This is only useful for
    very narrow cases.

    """
    if _should_use_sessions(request):
        language_code = request.session.get(translation.LANGUAGE_SESSION_KEY)
        if language_code:
            return language_code

    return request.COOKIES.get(settings.LANGUAGE_COOKIE_NAME)


def set_language(request, response, language_code):
    """Set the language in session or as a separate cookie.

    Sending language code as None removes the preference. response is not
    optional as Django 3.0 up always set the language cookie and Django 4.0
    will no longer set the language in the session.

    """
    if not language_code:

        if _should_use_sessions(request):
            try:
                del request.session[translation.LANGUAGE_SESSION_KEY]
            except KeyError:
                pass

        if response:
            try:
                response.delete_cookie(settings.LANGUAGE_COOKIE_NAME)
            except KeyError:
                pass

        return

    translation.activate(language_code)
    if _should_use_sessions(request):
        request.session[translation.LANGUAGE_SESSION_KEY] = language_code

    response.set_cookie(
        settings.LANGUAGE_COOKIE_NAME,
        language_code,
        max_age=settings.LANGUAGE_COOKIE_AGE,
        path=settings.LANGUAGE_COOKIE_PATH,
        domain=settings.LANGUAGE_COOKIE_DOMAIN,
    )



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for frontpage.
"""

from unittest.mock import patch

import pytest

from plinth.frontpage import Shortcut, add_custom_shortcuts

# pylint: disable=protected-access


@pytest.fixture(name='clean_global_shortcuts', autouse=True)
def fixture_clean_global_shortcuts():
    """Ensure that global list of shortcuts is clean."""
    Shortcut._all_shortcuts = {}


def test_shortcut_init_with_arguments():
    """Test initializing shortcut component without arguments."""
    with pytest.raises(ValueError):
        Shortcut(None, None)

    shortcut = Shortcut('test-component', 'test-name')
    assert shortcut.component_id == 'test-component'
    assert shortcut.name == 'test-name'
    assert shortcut.short_description is None
    assert shortcut.url == '?selected=test-component'
    assert shortcut.icon is None
    assert shortcut.description is None
    assert shortcut.manual_page is None
    assert shortcut.configure_url is None
    assert shortcut.clients is None
    assert not shortcut.login_required
    assert shortcut.allowed_groups is None
    assert Shortcut._all_shortcuts['test-component'] == shortcut


def test_shortcut_init():
    """Test initializing shortcut component."""
    clients = ['client1', 'client2']
    allowed_groups = ['group1', 'group2']
    shortcut = Shortcut('test-component', name='test-name',
                        short_description='test-short-description',
                        url='test-url', icon='test-icon',
                        description='test-description', manual_page='TestPage',
                        configure_url='test-configure-url', clients=clients,
                        login_required=True, allowed_groups=allowed_groups)
    assert shortcut.short_description == 'test-short-description'
    assert shortcut.url == 'test-url'
    assert shortcut.icon == 'test-icon'
    assert shortcut.description == 'test-description'
    assert shortcut.manual_page == 'TestPage'
    assert shortcut.configure_url == 'test-configure-url'
    assert shortcut.clients == clients
    assert shortcut.login_required
    assert shortcut.allowed_groups == set(allowed_groups)


def test_shortcut_remove():
    """Test removing a shortcut global list of shortcuts."""
    shortcut = Shortcut('test-component', None)
    shortcut.remove()
    with pytest.raises(KeyError):
        del Shortcut._all_shortcuts['test-component']


@pytest.fixture(name='common_shortcuts')
def fixture_common_shortcuts(clean_global_shortcuts):
    """Create some common shortcuts."""
    shortcuts = [
        Shortcut('anon-web-app-component-1', 'name1', 'short4', url='url1'),
        Shortcut('group1-web-app-component-1', 'Name2', 'Short3', url='url2',
                 login_required=True, allowed_groups=['group1']),
        Shortcut('group2-web-app-component-1', 'name3', 'short2', url='url3',
                 login_required=True, allowed_groups=['group2']),
        Shortcut('anon-non-web-app-component-1', 'name4', 'short1', url=None),
    ]
    return shortcuts


def test_shortcut_list_sorting(common_shortcuts):
    """Test listing shortcuts in sorted order."""
    cuts = common_shortcuts

    return_list = Shortcut.list()
    assert return_list == [cuts[0], cuts[1], cuts[2], cuts[3]]

    return_list = Shortcut.list(sort_by='name')
    assert return_list == [cuts[0], cuts[1], cuts[2], cuts[3]]

    return_list = Shortcut.list(sort_by='short_description')
    assert return_list == [cuts[3], cuts[2], cuts[1], cuts[0]]


def test_shortcut_list_web_apps_only(common_shortcuts):
    """Test listing only web app shortcuts."""
    cuts = common_shortcuts

    return_list = Shortcut.list()
    assert Shortcut.list() == [cuts[0], cuts[1], cuts[2], cuts[3]]

    return_list = Shortcut.list(web_apps_only=False)
    assert return_list == [cuts[0], cuts[1], cuts[2], cuts[3]]

    return_list = Shortcut.list(web_apps_only=True)
    assert return_list == [cuts[0], cuts[1], cuts[2]]


@patch('plinth.modules.users.privileged.get_user_groups')
def test_shortcut_list_with_username(get_user_groups, common_shortcuts):
    """Test listing for particular users."""
    cuts = common_shortcuts

    return_list = Shortcut.list()
    assert return_list == [cuts[0], cuts[1], cuts[2], cuts[3]]

    get_user_groups.return_value = ['admin']
    return_list = Shortcut.list(username='admin')
    assert return_list == [cuts[0], cuts[1], cuts[2], cuts[3]]

    get_user_groups.return_value = ['group1']
    return_list = Shortcut.list(username='user1')
    assert return_list == [cuts[0], cuts[1], cuts[3]]

    get_user_groups.return_value = ['group1', 'group2']
    return_list = Shortcut.list(username='user2')
    assert return_list == [cuts[0], cuts[1], cuts[2], cuts[3]]

    cut = Shortcut('group2-web-app-component-1', 'name5', 'short2', url='url4',
                   login_required=False, allowed_groups=['group3'])
    get_user_groups.return_value = ['group3']
    return_list = Shortcut.list(username='user3')
    assert return_list == [cuts[0], cuts[3], cut]

    get_user_groups.return_value = ['group4']
    return_list = Shortcut.list(username='user4')
    assert return_list == [cuts[0], cuts[3], cut]


def test_add_custom_shortcuts(shortcuts_file):
    """Test that adding custom shortcuts succeeds."""
    shortcuts_file('nextcloud.json')
    add_custom_shortcuts()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Forms for the Kiwix module.
"""

from django import forms
from django.core import validators
from django.utils.translation import gettext_lazy as _

from plinth import cfg
from plinth.utils import format_lazy

from .privileged import KIWIX_HOME


class AddPackageForm(forms.Form):
    """Form to upload a content package to a library."""

    # Would be nice to have a progress bar when uploading large files.
    file = forms.FileField(
        label=_('Upload File'), required=True, validators=[
            validators.FileExtensionValidator(
                ['zim'], _('Content packages have to be in .zim format'))
        ], help_text=format_lazy(
            _('Uploaded ZIM files will be stored under {kiwix_home}/content '
              'on your {box_name}. If Kiwix fails to add the file, it will be '
              'deleted immediately to save disk space.'),
            box_name=_(cfg.box_name), kiwix_home=KIWIX_HOME))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure or run unattended-upgrades."""

import logging
import os
import pathlib
import re
import subprocess
import time

from plinth.action_utils import (apt_hold, apt_hold_flag, apt_hold_freedombox,
                                 apt_unhold_freedombox, debconf_set_selections,
                                 is_package_manager_busy, run_apt_command,
                                 service_daemon_reload, service_is_running,
                                 service_restart, service_start, service_stop)
from plinth.actions import privileged
from plinth.modules.apache.components import check_url
from plinth.modules.snapshot import is_apt_snapshots_enabled
from plinth.modules.snapshot import is_supported as snapshot_is_supported
from plinth.modules.snapshot import load_augeas as snapshot_load_augeas

SOURCES_LIST = '/etc/apt/sources.list'
BACKPORTS_SOURCES_LIST = '/etc/apt/sources.list.d/freedombox2.list'

AUTO_CONF_FILE = '/etc/apt/apt.conf.d/20auto-upgrades'
LOG_FILE = '/var/log/unattended-upgrades/unattended-upgrades.log'
DPKG_LOG_FILE = '/var/log/unattended-upgrades/unattended-upgrades-dpkg.log'
RELEASE_FILE_URL = \
    'https://deb.debian.org/debian/dists/{}/Release'

APT_PREFERENCES_FREEDOMBOX = \
    '''Explanation: This file is managed by FreedomBox, do not edit.
Explanation: Allow carefully selected updates to 'freedombox' from backports.
Package: src:freedombox
Pin: release n={}-backports
Pin-Priority: 500
'''

# Whenever these preferences needs to change, increment the version number
# upgrades app. This ensures that setup is run again and the new contents are
# overwritten on the old file.
APT_PREFERENCES_APPS = \
    '''Explanation: This file is managed by FreedomBox, do not edit.
Explanation: matrix-synapse shall not be available in Debian stable but
Explanation: only in backports. Upgrade priority of packages that have needed
Explanation: versions only in backports.
Explanation: matrix-synapse >= 1.92.0-3 requires
Explanation: python3-canonicaljson >= 2.0.0~
Package: python3-canonicaljson
Pin: release n=bookworm-backports
Pin-Priority: 500
'''

DIST_UPGRADE_OBSOLETE_PACKAGES: list[str] = []

DIST_UPGRADE_PACKAGES_WITH_PROMPTS = [
    'bind9', 'firewalld', 'janus', 'minetest-server', 'minidlna',
    'mumble-server', 'radicale', 'roundcube-core', 'tt-rss'
]

DIST_UPGRADE_PRE_INSTALL_PACKAGES = ['base-files']

DIST_UPGRADE_PRE_DEBCONF_SELECTIONS: list[str] = [
    # Tell grub-pc to continue without installing grub again.
    'grub-pc grub-pc/install_devices_empty boolean true'
]

DIST_UPGRADE_REQUIRED_FREE_SPACE = 5000000

DIST_UPGRADE_SERVICE = '''
[Unit]
Description=Upgrade to new stable Debian release

[Service]
Type=oneshot
ExecStart=/usr/share/plinth/actions/actions upgrades dist_upgrade --no-args
KillMode=process
TimeoutSec=12hr
'''

DIST_UPGRADE_SERVICE_PATH = \
    '/run/systemd/system/freedombox-dist-upgrade.service'

dist_upgrade_flag = pathlib.Path(
    '/var/lib/freedombox/dist-upgrade-in-progress')


def _release_held_freedombox():
    """If freedombox package was left in held state, release it.

    This can happen due to an interrupted process.
    """
    if apt_hold_flag.exists() and not is_package_manager_busy():
        apt_unhold_freedombox()


@privileged
def run():
    """Run unattended-upgrades."""
    subprocess.run(['dpkg', '--configure', '-a'], check=False)
    run_apt_command(['--fix-broken', 'install'])
    _release_held_freedombox()

    subprocess.Popen(['systemctl', 'start', 'freedombox-manual-upgrade'],
                     stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL,
                     stderr=subprocess.DEVNULL, close_fds=True,
                     start_new_session=True)


def _check_auto() -> bool:
    """Check if automatic upgrades are enabled."""
    arguments = [
        'apt-config', 'shell', 'UpdateInterval',
        'APT::Periodic::Update-Package-Lists'
    ]
    output = subprocess.check_output(arguments).decode()
    update_interval = 0
    match = re.match(r"UpdateInterval='(.*)'", output)
    if match:
        update_interval = int(match.group(1))

    return bool(update_interval)


@privileged
def check_auto() -> bool:
    """Check if automatic upgrades are enabled."""
    return _check_auto()


@privileged
def enable_auto():
    """Enable automatic upgrades."""
    with open(AUTO_CONF_FILE, 'w', encoding='utf-8') as conffile:
        conffile.write('APT::Periodic::Update-Package-Lists "1";\n')
        conffile.write('APT::Periodic::Unattended-Upgrade "1";\n')


@privileged
def disable_auto():
    """Disable automatic upgrades."""
    with open(AUTO_CONF_FILE, 'w', encoding='utf-8') as conffile:
        conffile.write('APT::Periodic::Update-Package-Lists "0";\n')
        conffile.write('APT::Periodic::Unattended-Upgrade "0";\n')


@privileged
def get_log() -> str:
    """Return the automatic upgrades log."""
    log_lines = []
    try:
        log_lines.append('==> ' + os.path.basename(LOG_FILE))
        with open(LOG_FILE, 'r', encoding='utf-8') as file_handle:
            log_lines.append(file_handle.read())
    except IOError:
        pass

    try:
        log_lines.append('==> ' + os.path.basename(DPKG_LOG_FILE))
        with open(DPKG_LOG_FILE, 'r', encoding='utf-8') as file_handle:
            log_lines.append(file_handle.read())
    except IOError:
        pass

    return '\n'.join(log_lines)


def _get_protocol() -> str:
    """Return the protocol to use for newly added repository sources."""
    try:
        from plinth.modules.torproxy import utils
        if utils.is_apt_transport_tor_enabled():
            return 'tor+http'
    except Exception:
        pass

    return 'http'


def _is_release_file_available(protocol: str, dist: str,
                               backports=False) -> bool:
    """Return whether the release for dist[-backports] is available."""
    wrapper = None
    if protocol == 'tor+http':
        wrapper = 'torsocks'

    if backports:
        dist += '-backports'

    try:
        return check_url(RELEASE_FILE_URL.format(dist), wrapper=wrapper)
    except FileNotFoundError:
        return False


def _add_backports_sources(sources_list: str, protocol: str, dist: str):
    """Add backports sources to freedombox repositories list."""
    sources = '''# This file is managed by FreedomBox, do not edit.
# Allow carefully selected updates to 'freedombox' from backports.

deb {protocol}://deb.debian.org/debian {dist}-backports main
deb-src {protocol}://deb.debian.org/debian {dist}-backports main
'''
    sources = sources.format(protocol=protocol, dist=dist)
    with open(sources_list, 'w', encoding='utf-8') as file_handle:
        file_handle.write(sources)


def _check_and_backports_sources(develop=False):
    """Add backports sources after checking if it is available."""
    old_sources_list = '/etc/apt/sources.list.d/freedombox.list'
    if os.path.exists(old_sources_list):
        os.remove(old_sources_list)

    from plinth.modules.upgrades import (get_current_release,
                                         is_backports_current)
    if is_backports_current():
        logging.info('Repositories list up-to-date. Skipping update.')
        return

    try:
        with open('/etc/dpkg/origins/default', 'r',
                  encoding='utf-8') as default_origin:
            matches = [
                re.match(r'Vendor:\s+(Debian|FreedomBox)', line,
                         flags=re.IGNORECASE)
                for line in default_origin.readlines()
            ]
    except FileNotFoundError:
        logging.info('Could not open /etc/dpkg/origins/default')
        return

    if not any(matches):
        logging.info('System is running a derivative of Debian. Skip enabling '
                     'backports.')
        return

    release, dist = get_current_release()
    if release == 'unstable' or (release == 'testing' and not develop):
        logging.info(f'System release is {release}. Skip enabling backports.')
        return

    protocol = _get_protocol()
    if protocol == 'tor+http':
        logging.info('Package download over Tor is enabled.')

    if not _is_release_file_available(protocol, dist, backports=True):
        logging.info(
            f'Release file for {dist}-backports is not available yet.')
        return

    print(f'{dist}-backports is now available. Adding to sources.')
    _add_backports_sources(BACKPORTS_SOURCES_LIST, protocol, dist)
    # In case of dist upgrade, rewrite the preferences file.
    _add_apt_preferences()


def _add_apt_preferences():
    """Setup APT preferences to upgrade selected packages from backports."""
    base_path = pathlib.Path('/etc/apt/preferences.d')
    for file_name in ['50freedombox.pref', '50freedombox2.pref']:
        full_path = base_path / file_name
        if full_path.exists():
            full_path.unlink()

    # Don't try to remove 50freedombox3.pref as this file is shipped with the
    # Debian package and is removed using maintainer scripts.

    from plinth.modules.upgrades import get_current_release
    _, dist = get_current_release()
    if dist == 'sid':
        logging.info(
            f'System distribution is {dist}. Skip setting apt preferences '
            'for backports.')
    else:
        logging.info(f'Setting apt preferences for {dist}-backports.')
        with open(base_path / '50freedombox4.pref', 'w',
                  encoding='utf-8') as file_handle:
            file_handle.write(APT_PREFERENCES_FREEDOMBOX.format(dist))
        with open(base_path / '51freedombox-apps.pref', 'w',
                  encoding='utf-8') as file_handle:
            file_handle.write(APT_PREFERENCES_APPS)


def _is_sufficient_free_space() -> bool:
    """Return whether there is sufficient free space for dist upgrade."""
    output = subprocess.check_output(['df', '--output=avail', '/'])
    free_space = int(output.decode().split('\n')[1])
    return free_space >= DIST_UPGRADE_REQUIRED_FREE_SPACE


def _check_dist_upgrade(test_upgrade=False) -> tuple[bool, str]:
    """Check if a distribution upgrade be performed.

    Check for new stable release, if updates are enabled, and if there is
    enough free space for the dist upgrade.

    If test_upgrade is True, also check for upgrade to testing.

    Return (boolean, string) indicating if the upgrade is ready, and a reason
    if not.
    """
    if dist_upgrade_flag.exists():
        return (True, 'found-previous')

    from plinth.modules.upgrades import get_current_release
    release, dist = get_current_release()
    if release in ['unstable', 'testing']:
        return (False, f'already-{release}')

    check_dists = ['stable']
    if test_upgrade:
        check_dists.append('testing')

    codename = None
    for check_dist in check_dists:
        url = RELEASE_FILE_URL.format(check_dist)
        command = ['curl', '--silent', '--location', '--fail', url]
        protocol = _get_protocol()
        if protocol == 'tor+http':
            command.insert(0, 'torsocks')
            logging.info('Package download over Tor is enabled.')

        try:
            output = subprocess.check_output(command).decode()
        except (subprocess.CalledProcessError, FileNotFoundError):
            logging.warning('Error while checking for new %s release',
                            check_dist)
        else:
            for line in output.split('\n'):
                if line.startswith('Codename:'):
                    codename = line.split()[1]

    if not codename:
        return (False, 'codename-not-found')

    if codename == dist:
        return (False, f'already-{dist}')

    if not _check_auto():
        return (False, 'upgrades-not-enabled')

    if check_dist == 'testing' and not test_upgrade:
        return (False, 'test-not-set')

    if not _is_sufficient_free_space():
        return (False, 'not-enough-free-space')

    logging.info('Upgrading from %s to %s...', dist, codename)
    with open(SOURCES_LIST, 'r', encoding='utf-8') as sources_list:
        lines = sources_list.readlines()

    with open(SOURCES_LIST, 'w', encoding='utf-8') as sources_list:
        for line in lines:
            # E.g. replace 'bullseye' with 'bookworm'.
            new_line = line.replace(dist, codename)
            if check_dist == 'testing':
                # E.g. replace 'stable' with 'bookworm'.
                new_line = new_line.replace('stable', codename)

            sources_list.write(new_line)

    logging.info('Dist upgrade in progress. Setting flag.')
    dist_upgrade_flag.touch(mode=0o660)
    return (True, 'started-dist-upgrade')


def _take_snapshot_and_disable() -> bool:
    """Take a snapshot if supported and enabled, then disable snapshots.

    Return whether snapshots shall be re-enabled at the end.
    """
    if snapshot_is_supported():
        print('Taking a snapshot before dist upgrade...', flush=True)
        subprocess.run([
            '/usr/share/plinth/actions/actions', 'snapshot', 'create',
            '--no-args'
        ], check=True)
        aug = snapshot_load_augeas()
        if is_apt_snapshots_enabled(aug):
            print('Disable apt snapshots during dist upgrade...', flush=True)
            subprocess.run([
                '/usr/share/plinth/actions/actions',
                'snapshot',
                'disable_apt_snapshot',
            ], input='{"args": ["yes"], "kwargs": {}}'.encode(), check=True)
            return True
        else:
            print('Apt snapshots already disabled.', flush=True)
    else:
        print('Snapshots are not supported, skip taking a snapshot.',
              flush=True)

    return False


def _restore_snapshots_config(reenable=False):
    """Restore original snapshots configuration."""
    if reenable:
        print('Re-enable apt snapshots...', flush=True)
        subprocess.run([
            '/usr/share/plinth/actions/actions', 'snapshot',
            'disable_apt_snapshot'
        ], input='{"args": ["no"], "kwargs": {}}'.encode(), check=True)


def _disable_searx() -> bool:
    """If searx is enabled, disable it until we can upgrade it properly.

    Return whether searx was originally enabled.
    """
    searx_is_enabled = pathlib.Path(
        '/etc/uwsgi/apps-enabled/searx.ini').exists()
    if searx_is_enabled:
        print('Disabling searx...', flush=True)
        subprocess.run(
            ['/usr/share/plinth/actions/actions', 'apache', 'uwsgi_disable'],
            input='{"args": ["searx"], "kwargs": {}}'.encode(), check=True)

    return searx_is_enabled


def _update_searx(reenable=False):
    """If searx is installed, update search engines list.

    Re-enable if previously enabled.
    """
    if pathlib.Path('/etc/searx/settings.yml').exists():
        print('Updating searx search engines list...', flush=True)
        subprocess.run([
            '/usr/share/plinth/actions/actions', 'searx', 'setup', '--no-args'
        ], check=True)
        if reenable:
            print('Re-enabling searx after upgrade...', flush=True)
            subprocess.run([
                '/usr/share/plinth/actions/actions', 'apache', 'uwsgi_enable'
            ], input='{"args": ["searx"], "kwargs": {}}'.encode(), check=True)


def _perform_dist_upgrade():
    """Perform upgrade to next release of Debian."""
    reenable_snapshots = _take_snapshot_and_disable()
    reenable_searx = _disable_searx()

    # If quassel is running during dist upgrade, it may be restarted
    # several times. This causes IRC users to rapidly leave/join
    # channels. Stop quassel for the duration of the dist upgrade.
    quassel_service = 'quasselcore'
    quassel_was_running = service_is_running(quassel_service)
    if quassel_was_running:
        print('Stopping quassel service before dist upgrade...', flush=True)
        service_stop(quassel_service)

    # Hold freedombox package during entire dist upgrade.
    print('Holding freedombox package...', flush=True)
    with apt_hold_freedombox():
        print('Updating Apt cache...', flush=True)
        run_apt_command(['update'])

        # Install packages that are necessary for unattended-upgrades
        # to start the dist upgrade.
        print(f'Upgrading packages: {DIST_UPGRADE_PRE_INSTALL_PACKAGES}...',
              flush=True)
        run_apt_command(['install'] + DIST_UPGRADE_PRE_INSTALL_PACKAGES)

        # Pre-set debconf selections if they are required during the
        # dist upgrade.
        if DIST_UPGRADE_PRE_DEBCONF_SELECTIONS:
            print(
                f'Setting debconf selections: '
                f'{DIST_UPGRADE_PRE_DEBCONF_SELECTIONS}', flush=True)
            debconf_set_selections(DIST_UPGRADE_PRE_DEBCONF_SELECTIONS)

        # This will upgrade most of the packages.
        # Previously, when dist-upgrading from bullseye to bookworm, there was
        # an issue where unattended-upgrade gets stuck. See #2266.  However, it
        # does not get stuck when dist-upgrading from bookworm to trixie.
        print('Running unattended-upgrade...', flush=True)
        subprocess.run(['unattended-upgrade', '--verbose'], check=False)

        # Remove obsolete packages that may prevent other packages from
        # upgrading.
        if DIST_UPGRADE_OBSOLETE_PACKAGES:
            print(f'Removing packages: {DIST_UPGRADE_OBSOLETE_PACKAGES}...',
                  flush=True)
            run_apt_command(['remove'] + DIST_UPGRADE_OBSOLETE_PACKAGES)

        # Hold packages known to have conffile prompts. FreedomBox service
        # will handle their upgrade later.
        print(
            'Holding packages with conffile prompts: ' +
            ', '.join(DIST_UPGRADE_PACKAGES_WITH_PROMPTS) + '...', flush=True)
        with apt_hold(DIST_UPGRADE_PACKAGES_WITH_PROMPTS):
            print('Running apt full-upgrade...', flush=True)
            returncode = run_apt_command(['full-upgrade'])

        # Check if apt upgrade was successful.
        if returncode:
            raise RuntimeError(
                'Apt full-upgrade was not successful. Distribution upgrade '
                'will be retried at a later time.')

        _update_searx(reenable_searx)

        if quassel_was_running:
            print('Re-starting quassel service after dist upgrade...',
                  flush=True)
            service_start(quassel_service)

        print('Running apt autoremove...', flush=True)
        run_apt_command(['autoremove'])

    # Run unattended-upgrade once more to handle upgrading the
    # freedombox package.
    print('Running unattended-upgrade...', flush=True)
    subprocess.run(['unattended-upgrade', '--verbose'], check=False)

    _restore_snapshots_config(reenable_snapshots)

    # Restart FreedomBox service to ensure it is using the latest
    # dependencies.
    print('Restarting FreedomBox service...', flush=True)
    service_restart('plinth')

    # After 10 minutes, update apt cache again to trigger force_upgrades.
    print('Waiting for 10 minutes...', flush=True)
    time.sleep(10 * 60)
    print('Updating Apt cache...', flush=True)
    run_apt_command(['update'])

    print('Dist upgrade complete. Removing flag.', flush=True)
    if dist_upgrade_flag.exists():
        dist_upgrade_flag.unlink()


@privileged
def setup():
    """Setup apt preferences."""
    _add_apt_preferences()


@privileged
def activate_backports(develop: bool = False):
    """Setup software repositories needed for FreedomBox.

    Repositories list for now only contains the backports. If the file exists,
    assume that it contains backports.
    """
    _check_and_backports_sources(develop)


def _start_dist_upgrade_service():
    """Create dist upgrade service and start it."""
    with open(DIST_UPGRADE_SERVICE_PATH, 'w',
              encoding='utf-8') as service_file:
        service_file.write(DIST_UPGRADE_SERVICE)

    service_daemon_reload()
    subprocess.Popen(['systemctl', 'start', 'freedombox-dist-upgrade'],
                     stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL,
                     stderr=subprocess.DEVNULL, close_fds=True,
                     start_new_session=True)


@privileged
def start_dist_upgrade(test: bool = False) -> dict[str, str | bool]:
    """Start dist upgrade process.

    Check if a new stable release is available, and start dist-upgrade process
    if updates are enabled.
    """
    _release_held_freedombox()

    upgrade_ready, reason = _check_dist_upgrade(test)
    if upgrade_ready:
        _start_dist_upgrade_service()

    return {'dist_upgrade_started': upgrade_ready, 'reason': reason}


@privileged
def dist_upgrade():
    """Perform major distribution upgrade."""
    _perform_dist_upgrade()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app to configure name services.
"""

import logging

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, menu
from plinth.modules.backups.components import BackupRestore
from plinth.signals import domain_added, domain_removed
from plinth.utils import format_lazy

from . import components, manifest

logger = logging.getLogger(__name__)

_description = [
    format_lazy(
        _('Name Services provides an overview of the ways {box_name} can be '
          'reached from the public Internet: domain name, Tor onion service, '
          'and Pagekite. For each type of name, it is shown whether the HTTP, '
          'HTTPS, and SSH services are enabled or disabled for incoming '
          'connections through the given name.'), box_name=(cfg.box_name))
]


class NamesApp(app_module.App):
    """FreedomBox app for names."""

    app_id = 'names'

    _version = 1

    can_be_disabled = False

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()
        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True, name=_('Name Services'),
                               icon='fa-tags', description=_description,
                               manual_page='NameServices')
        self.add(info)

        menu_item = menu.Menu('menu-names', info.name, None, info.icon,
                              'names:index',
                              parent_url_name='system:visibility', order=10)
        self.add(menu_item)

        backup_restore = BackupRestore('backup-restore-names',
                                       **manifest.backup)
        self.add(backup_restore)

    @staticmethod
    def post_init():
        """Perform post initialization operations."""
        domain_added.connect(on_domain_added)
        domain_removed.connect(on_domain_removed)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        self.enable()


def on_domain_added(sender, domain_type, name='', description='',
                    services=None, **kwargs):
    """Add domain to global list."""
    if not domain_type:
        return

    if not name:
        return
    if not services:
        services = []

    components.DomainName('domain-' + sender + '-' + name, name, domain_type,
                          services)
    logger.info('Added domain %s of type %s with services %s', name,
                domain_type, str(services))


def on_domain_removed(sender, domain_type, name='', **kwargs):
    """Remove domain from global list."""
    if name:
        component_id = 'domain-' + sender + '-' + name
        components.DomainName.get(component_id).remove()
        logger.info('Removed domain %s of type %s', name, domain_type)
    else:
        for domain_name in components.DomainName.list():
            if domain_name.domain_type.component_id == domain_type:
                domain_name.remove()

                logger.info('Remove domain %s of type %s', domain_name.name,
                            domain_type)


######################################################
# Domain utilities meant to be used by other modules #
######################################################


def get_available_tls_domains():
    """Return an iterator with all domains able to have a certificate."""
    return (domain.name for domain in components.DomainName.list()
            if domain.domain_type.can_have_certificate)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for api for android app.
"""

from plinth import app as app_module


class ApiApp(app_module.App):
    """FreedomBox app for API for Android app."""

    app_id = 'api'

    _version = 1

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True)
        self.add(info)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        self.enable()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Configure Cockpit.
"""

import augeas

from plinth import action_utils
from plinth.actions import privileged

CONFIG_FILE = '/etc/cockpit/cockpit.conf'


def _load_augeas():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/inifile/lens', 'Puppet.lns')
    aug.set('/augeas/load/inifile/incl[last() + 1]', CONFIG_FILE)
    aug.load()
    return aug


@privileged
def setup():
    """Setup Cockpit configuration."""
    aug = _load_augeas()
    aug.set('/files' + CONFIG_FILE + '/WebService/UrlRoot', '/_cockpit/')
    aug.remove('/files' + CONFIG_FILE + '/WebService/Origins')
    aug.save()
    action_utils.service_restart('cockpit.socket')
    # Accommodate changes in Apache configuration file from v1 to v2.
    action_utils.service_reload('apache2')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the snapshot module.
"""

from django.urls import re_path

from . import views

urlpatterns = [
    re_path(r'^sys/snapshot/$', views.SnapshotAppView.as_view(), name='index'),
    re_path(r'^sys/snapshot/manage/$', views.manage, name='manage'),
    re_path(r'^sys/snapshot/selected/delete$', views.delete_selected,
            name='delete-selected'),
    re_path(r'^sys/snapshot/(?P<number>\d+)/rollback$', views.rollback,
            name='rollback'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure RSS-Bridge."""

import pathlib

from plinth import action_utils
from plinth.actions import privileged

PUBLIC_ACCESS_FILE = pathlib.Path('/etc/rss-bridge/is_public')
ENABLE_LIST = pathlib.Path('/etc/rss-bridge/whitelist.txt')


@privileged
def setup():
    """Configure RSS-Bridge by enable all bridges."""
    ENABLE_LIST.write_text('*\n', encoding='utf-8')


@privileged
def set_public(enable: bool):
    """Allow/disallow public access."""
    if enable:
        PUBLIC_ACCESS_FILE.touch()
    else:
        PUBLIC_ACCESS_FILE.unlink(missing_ok=True)

    action_utils.service_reload('apache2')


def is_public() -> bool:
    """Return whether public access is enabled."""
    return PUBLIC_ACCESS_FILE.exists()


@privileged
def uninstall():
    """Remove config files when app is uninstalled."""
    for path in PUBLIC_ACCESS_FILE, ENABLE_LIST:
        path.unlink(missing_ok=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Forms for configuring Tor Proxy.
"""

from django import forms
from django.utils.translation import gettext_lazy as _

from plinth.modules.tor.forms import TorCommonForm


class TorProxyForm(TorCommonForm):
    """Tor Proxy configuration form."""
    apt_transport_tor_enabled = forms.BooleanField(
        label=_('Download software packages over Tor'), required=False,
        help_text=_('When enabled, software will be downloaded over the Tor '
                    'network for installations and upgrades. This adds a '
                    'degree of privacy and security during software '
                    'downloads.'))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure MediaWiki."""

import pathlib
from urllib.parse import urlparse

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import frontpage, menu
from plinth.config import DropinConfigs
from plinth.daemon import Daemon
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.package import Packages

from . import manifest, privileged

_description = [
    _('MediaWiki is the wiki engine that powers Wikipedia and other WikiMedia '
      'projects. A wiki engine is a program for creating a collaboratively '
      'edited website. You can use MediaWiki to host a wiki-like website, '
      'take notes or collaborate with friends on projects.'),
    _('This MediaWiki instance comes with a randomly generated administrator '
      'password. You can set a new password in the "Configuration" section '
      'and log in using the "admin" account. You can then create more user '
      'accounts from MediaWiki itself by going to the <a '
      'href="/mediawiki/index.php/Special:CreateAccount">'
      'Special:CreateAccount</a> page.'),
    _('Anyone with a link to this wiki can read it. Only users that are '
      'logged in can make changes to the content.')
]

USER_CONFIG_FILE = '/etc/mediawiki/FreedomBoxSettings.php'


class MediaWikiApp(app_module.App):
    """FreedomBox app for MediaWiki."""

    app_id = 'mediawiki'

    _version = 12

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()
        self._private_mode = True

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('MediaWiki'), icon_filename='mediawiki',
                               short_description=_('Wiki'),
                               description=_description,
                               manual_page='MediaWiki',
                               clients=manifest.clients)
        self.add(info)

        menu_item = menu.Menu('menu-mediawiki', info.name,
                              info.short_description, info.icon_filename,
                              'mediawiki:index', parent_url_name='apps')
        self.add(menu_item)

        shortcut = Shortcut('shortcut-mediawiki', info.name,
                            short_description=info.short_description,
                            icon=info.icon_filename, url='/mediawiki',
                            clients=info.clients, login_required=True)
        self.add(shortcut)

        packages = Packages('packages-mediawiki',
                            ['mediawiki', 'imagemagick', 'php-sqlite3'])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-mediawiki', [
            '/etc/apache2/conf-available/mediawiki-freedombox.conf',
            '/etc/mediawiki/FreedomBoxStaticSettings.php',
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-mediawiki', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        webserver = Webserver('webserver-mediawiki', 'mediawiki',
                              urls=['https://{host}/mediawiki/'],
                              expect_redirects=True)
        self.add(webserver)

        webserver = Webserver('webserver-mediawiki-freedombox',
                              'mediawiki-freedombox')
        self.add(webserver)

        daemon = Daemon('daemon-mediawiki', 'mediawiki-jobrunner')
        self.add(daemon)

        backup_restore = BackupRestore('backup-restore-mediawiki',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        privileged.update()
        if not old_version:
            self.enable()

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()


class Shortcut(frontpage.Shortcut):
    """Frontpage shortcut for only logged users when in private mode."""

    def enable(self):
        """When enabled, check if MediaWiki is in private mode."""
        super().enable()
        self.login_required = get_config()['enable_private_mode']


def _get_config_value_in_file(setting_name, config_file):
    """Return the value of a setting from a config file."""
    with open(config_file, 'r', encoding='utf-8') as config:
        for line in config:
            if line.startswith(setting_name):
                return line.partition('=')[2].strip('\n ;\'"')

        return None


def _get_static_config_file():
    """Return the path for the file containing static settings."""
    base_path = ('/usr/share/freedombox/etc/'
                 'mediawiki/FreedomBoxStaticSettings.php')
    for path in [
            pathlib.Path(base_path),
            pathlib.Path(__file__).parent / 'data' / base_path.lstrip('/')
    ]:
        if path.exists():
            return path

    raise RuntimeError('Unable to find static config file')


def _get_config_value(setting_name):
    """Return a configuration value from multiple configuration files."""
    return _get_config_value_in_file(setting_name, USER_CONFIG_FILE) or \
        _get_config_value_in_file(setting_name, _get_static_config_file())


def get_config():
    """Return all the configuration settings."""
    server_url = _get_config_value('$wgServer')
    create_permission = _get_config_value(
        "$wgGroupPermissions['*']['createaccount']")
    read_permission = _get_config_value("$wgGroupPermissions['*']['read']")
    return {
        'default_skin': _get_config_value('$wgDefaultSkin'),
        'domain': urlparse(server_url).netloc,
        'site_name': _get_config_value('$wgSitename') or 'Wiki',
        'default_lang': _get_config_value('$wgLanguageCode') or None,
        'enable_public_registrations': 'true' in create_permission,
        'enable_private_mode': 'false' in read_permission,
    }


def set_server_url(domain):
    """Set the value of $wgServer."""
    protocol = 'https'
    if domain.endswith('.onion'):
        protocol = 'http'

    privileged.set_server_url(f'{protocol}://{domain}')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Utilities for functional testing.
"""

import configparser
import logging
import os
import pathlib
import subprocess
import tempfile
import time
import warnings
from contextlib import contextmanager

import pytest
import requests
from selenium.common.exceptions import (StaleElementReferenceException,
                                        WebDriverException)
from selenium.webdriver.support.ui import WebDriverWait

config = configparser.ConfigParser()
config.read(pathlib.Path(__file__).with_name('config.ini'))

# Configuration to allow each pytest-xdist worker to hit a dedicated
# app server. See .ci/functional-tests.yml for usage.
worker = os.environ.get('PYTEST_XDIST_WORKER', 'master')
if worker == 'master':
    config['DEFAULT']['url'] = os.environ.get(
        'FREEDOMBOX_URL', config['DEFAULT']['url']).rstrip('/')
else:
    # worker_ids are like gw0, gw1, ...
    worker_number = int(worker.lstrip('gw')) + 1
    config['DEFAULT']['url'] = os.environ[f'APP_SERVER_URL_{worker_number}']

config['DEFAULT']['ssh_port'] = os.environ.get('FREEDOMBOX_SSH_PORT',
                                               config['DEFAULT']['ssh_port'])
config['DEFAULT']['samba_port'] = os.environ.get(
    'FREEDOMBOX_SAMBA_PORT', config['DEFAULT']['samba_port'])

logger = logging.getLogger(__name__)

base_url = config['DEFAULT']['url']

# unlisted sites just use '/' + site_name as url
_site_url = {
    'wiki': '/ikiwiki',
    'jsxc': '/plinth/apps/jsxc/jsxc/',
    'cockpit': '/_cockpit/',
    'syncthing': '/syncthing/',
    'rssbridge': '/rss-bridge/',
    'ttrss': '/tt-rss/',
}

_sys_modules = [
    'avahi', 'backups', 'bind', 'cockpit', 'config', 'datetime', 'diagnostics',
    'dynamicdns', 'firewall', 'letsencrypt', 'names', 'networks', 'pagekite',
    'performance', 'power', 'privacy', 'security', 'snapshot', 'ssh',
    'storage', 'upgrades', 'users'
]


######################
# Browser Extensions #
######################
def visit(browser, path):
    """Visit a path assuming the base URL as configured."""
    browser.visit(config['DEFAULT']['url'] + path)


def eventually(function, args=[], timeout=30):
    """Execute a function returning a boolean expression till it returns
    True or a timeout is reached"""
    end_time = time.time() + timeout
    current_time = time.time()
    while current_time < end_time:
        try:
            if function(*args):
                return True
        except Exception:
            pass

        time.sleep(0.1)
        current_time = time.time()

    return False


class _PageLoaded():
    """
    Wait until a page (re)loaded.

    - element: Wait until this element gets stale
    - expected_url (optional): Wait for the URL to become <expected_url>.
      This can be necessary to wait for a redirect to finish.
    """

    def __init__(self, element, expected_url=None):
        self.element = element
        self.expected_url = expected_url

    def __call__(self, driver):
        is_stale = False
        try:
            self.element.has_class('whatever_class')
        except (StaleElementReferenceException, TypeError):
            # After a domain name change, Let's Encrypt will restart the web
            # server and could cause a connection failure.
            if driver.find_by_id('netErrorButtonContainer'):
                try:
                    driver.visit(driver.url)
                except WebDriverException:
                    pass
                return False

            is_fully_loaded = driver.execute_script(
                'return document.readyState;') == 'complete'
            if not is_fully_loaded:
                is_stale = False
            elif self.expected_url is None:
                is_stale = True
            else:
                if driver.url.endswith(self.expected_url):
                    is_stale = True

        return is_stale


@contextmanager
def wait_for_page_update(browser, timeout=300, expected_url=None):
    page_body = browser.find_by_tag('body').first
    try:
        yield
    except WebDriverException:
        # ignore a connection failure which may happen after web server restart
        pass

    WebDriverWait(browser, timeout).until(_PageLoaded(page_body, expected_url))


def _get_site_url(site_name):
    if site_name.startswith('share'):
        site_name = site_name.replace('_', '/')
    url = '/' + site_name
    url = _site_url.get(site_name, url)
    return url


def access_url(browser, site_name):
    browser.visit(config['DEFAULT']['url'] + _get_site_url(site_name))


def get_password(username):
    """Get a user password."""
    if username == config['DEFAULT']['username']:
        return config['DEFAULT']['password']
    return 'password@&123_{}'.format(username)


def is_available(browser, site_name):
    """Check if the given site_name is available."""
    url_to_visit = config['DEFAULT']['url'] + _get_site_url(site_name)
    browser.visit(url_to_visit)
    time.sleep(3)
    browser.reload()
    if '404' in browser.title or 'Page not found' in browser.title:
        return False

    # The site might have a default path after the sitename,
    # e.g /mediawiki/Main_Page
    print('URL =', browser.url, url_to_visit, browser.title)
    browser_url = browser.url.partition('://')[2]
    url_to_visit_without_proto = url_to_visit.strip('/').partition('://')[2]
    return browser_url.startswith(url_to_visit_without_proto)  # not a redirect


def download_file(browser, url):
    """Return file contents after downloading a URL."""
    cookies = browser.cookies.all()
    response = requests.get(url, cookies=cookies, verify=False)
    if response.status_code != 200:
        raise Exception('URL download failed')

    return response.content


def download_file_outside_browser(url):
    """Download a file to disk given a URL."""
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        logging.captureWarnings(True)
        request = requests.get(url, verify=False)
        logging.captureWarnings(False)
        temp_file.write(request.content)

    return temp_file.name


###########################
# Form handling utilities #
###########################
def submit(browser, element=None, form_class=None, expected_url=None):
    """Submit a specific form in the current page and wait for page change."""
    if not (element or form_class):
        raise AssertionError('Either element or form_class must be sent')

    with wait_for_page_update(browser, expected_url=expected_url):
        if element:
            element.click()
        elif form_class:
            browser.find_by_css(
                '.{} input[type=submit]'.format(form_class)).click()
        else:
            browser.find_by_css('input[type=submit]').click()


def set_app_form_value(browser, app_id, element_id, value):
    """Change a value in the form and submit."""
    nav_to_module(browser, app_id)
    original_url = browser.url
    browser.find_by_id(element_id).fill(value)
    submit(browser, form_class='form-configuration')
    # Check that there are no errors
    assert browser.url == original_url
    assert not browser.find_by_css('.alert-danger')


def change_checkbox_status(browser, app_name, checkbox_id,
                           change_status_to='enabled'):
    """Change checkbox status."""
    checkbox = browser.find_by_id(checkbox_id)
    if change_status_to == 'enabled':
        checkbox.check()
    else:
        checkbox.uncheck()

    submit(browser, form_class='form-configuration')


def wait_for_config_update(browser, app_name):
    """Wait until the configuration update progress goes away.

    Perform an atomic check that page is fully loaded and that progress icon is
    not present at the same time to avoid race conditions due to automatic page
    reloads.

    """
    script = 'return (document.readyState == "complete") && ' \
        '(!Boolean(document.querySelector(".app-operation")));'
    while not browser.execute_script(script):
        time.sleep(0.1)


############################
# Login handling utilities #
############################
def is_login_prompt(browser):
    return all(
        [browser.find_by_id('id_username'),
         browser.find_by_id('id_password')])


def _create_admin_account(browser, username, password):
    browser.find_by_id('id_username').fill(username)
    browser.find_by_id('id_password1').fill(password)
    browser.find_by_id('id_password2').fill(password)
    submit(browser, form_class='form-create')


def login(browser):
    """Login to the FreedomBox interface with default test account."""
    login_with_account(browser, base_url, config['DEFAULT']['username'],
                       config['DEFAULT']['password'])


def login_with_account(browser, url, username, password=None):
    """Login to the FreedomBox interface with provided account."""
    if password is None:
        password = get_password(username)
    # XXX: Find a way to remove the hardcoded jsxc URL
    if '/plinth/' not in browser.url or '/jsxc/jsxc' in browser.url:
        browser.visit(url)

    user_menu = browser.find_by_id('id_user_menu')

    if len(user_menu):
        if user_menu.text == username:
            return

        logout(browser)

    login_button = browser.links.find_by_href('/plinth/accounts/login/')
    if login_button:
        login_button.first.click()
        if login_button:
            browser.fill('username', username)
            browser.fill('password', password)
            submit(browser, form_class='form-login')
    else:
        browser.visit(base_url + '/plinth/firstboot/welcome')
        submit(browser, form_class='form-start')  # "Start Setup" button
        _create_admin_account(browser, username, password)
        if '/network-topology-first-boot' in browser.url:
            submit(browser, element=browser.find_by_name('skip')[0])

        if '/internet-connection-type' in browser.url:
            submit(browser, element=browser.find_by_name('skip')[0])

        if '/firstboot/backports' in browser.url:
            submit(browser, element=browser.find_by_name('next')[0])

        if '/firstboot/update' in browser.url:
            browser.find_by_id('id_update_now').uncheck()
            submit(browser, element=browser.find_by_name('next')[0])


def logout(browser):
    """Log out of the FreedomBox interface."""
    # Navigate to the home page if logout form is not found
    if not browser.find_by_css('.form-logout'):
        visit(browser, '/plinth/')

    # We are not logged in if the home page does not contain logout form
    if browser.find_by_css('.form-logout'):
        browser.find_by_id('id_user_menu').click()
        submit(browser, form_class='form-logout')


#################
# App utilities #
#################
def nav_to_module(browser, module):
    sys_or_apps = 'sys' if module in _sys_modules else 'apps'
    required_url = base_url + f'/plinth/{sys_or_apps}/{module}/'
    if browser.url != required_url:
        browser.visit(required_url)


def app_select_domain_name(browser, app_name, domain_name):
    browser.visit('{}/plinth/apps/{}/setup/'.format(base_url, app_name))
    drop_down = browser.find_by_id('id_domain_name')
    drop_down.select(domain_name)
    submit(browser, form_class='form-configuration')


#########################
# App install utilities #
#########################
def is_installed(browser, app_name):
    nav_to_module(browser, app_name)
    return not bool(browser.find_by_css('.form-install input[type=submit]'))


def install(browser, app_name):
    nav_to_module(browser, app_name)
    install_button_css = '.form-install input[type=submit]'
    while True:
        script = 'return (document.readyState == "complete") && ' \
            '(!Boolean(document.querySelector(".app-operation"))) &&' \
            '(!Boolean(document.querySelector(".app-just-installed")));'
        if not browser.execute_script(script):
            time.sleep(0.1)
        elif browser.is_element_present_by_css('.neterror'):
            browser.visit(browser.url)
        elif browser.is_element_present_by_css('.alert-danger'):
            break
        elif browser.is_element_present_by_css(install_button_css):
            install_button = browser.find_by_css(install_button_css).first
            if install_button['disabled']:
                # This app is not available in this distribution
                warnings.warn(
                    f'App {app_name} is not available in distribution')
                pytest.skip('App not available in distribution')
            else:
                install_button.click()
        else:
            break


def uninstall(browser, app_name):
    """Uninstall the app if possible."""
    nav_to_module(browser, app_name)
    actions_button = browser.find_by_id('id_extra_actions_button')
    if not actions_button:
        pytest.skip('App cannot be uninstalled')

    actions_button.click()
    uninstall_item = browser.find_by_css('.uninstall-item')
    if not uninstall_item:
        pytest.skip('App cannot be uninstalled')

    uninstall_item[0].click()
    submit(browser, form_class='form-uninstall')

    while True:
        script = 'return (document.readyState == "complete") && ' \
            '(!Boolean(document.querySelector(".app-operation")));'
        if not browser.execute_script(script):
            time.sleep(0.1)
        elif browser.is_element_present_by_css('.neterror'):
            browser.visit(browser.url)
        elif browser.is_element_present_by_css('.alert-danger'):
            raise RuntimeError('Uninstall failed')
        else:
            break


################################
# App enable/disable utilities #
################################
def _change_app_status(browser, app_name, change_status_to='enabled'):
    """Enable or disable application."""
    button = browser.find_by_css('button[name="app_enable_disable_button"]')
    if not button:
        raise RuntimeError('App enable/disable button not found')

    should_enable_field = browser.find_by_id('id_should_enable')
    if (should_enable_field.value == 'False' and change_status_to
            == 'disabled') or (should_enable_field.value == 'True'
                               and change_status_to == 'enabled'):
        submit(browser, element=button)


def app_enable(browser, app_name):
    nav_to_module(browser, app_name)
    _change_app_status(browser, app_name, 'enabled')


def app_disable(browser, app_name):
    nav_to_module(browser, app_name)
    _change_app_status(browser, app_name, 'disabled')


def app_is_enabled(browser, app_name):
    """Return whether the app is enabled."""
    nav_to_module(browser, app_name)
    should_enable_field = browser.find_by_id('id_should_enable')
    return should_enable_field.value == 'False'


def app_can_be_disabled(browser, app_name):
    """Return whether the application can be disabled."""
    nav_to_module(browser, app_name)
    button = browser.find_by_css('button[name="app_enable_disable_button"]')
    return bool(button)


#########################
# Domain name utilities #
#########################
def set_domain_name(browser, domain_name):
    nav_to_module(browser, 'config')
    browser.find_by_id('id_domainname').fill(domain_name)
    submit(browser, form_class='form-configuration')


########################
# Front page utilities #
########################
def find_on_front_page(browser, app_name):
    browser.visit(base_url)
    shortcuts = browser.links.find_by_href(f'/{app_name}/')
    return shortcuts


def is_visible_on_front_page(browser, app_name):
    shortcuts = find_on_front_page(browser, app_name)
    return len(shortcuts) == 1


####################
# Daemon utilities #
####################
def service_is_running(browser, app_name):
    nav_to_module(browser, app_name)
    return len(browser.find_by_id('service-not-running')) == 0


def service_is_not_running(browser, app_name):
    nav_to_module(browser, app_name)
    return len(browser.find_by_id('service-not-running')) != 0


def running_inside_container():
    """Check if freedombox is running inside a container"""
    # If the URL to connect to was overridden then assume that we are running
    # tests on a different machine than the machine running freedombox. Assume
    # running inside container to be conservative about tests.
    if config['DEFAULT']['url'] != 'https://localhost':
        return True

    # If URL is not overridden then testing code and freedombox are running on
    # the same machine. Proceed with a proper test.
    result = subprocess.run(['systemd-detect-virt', '--container'],
                            stdout=subprocess.PIPE, check=False)
    return result.stdout.decode('utf-8').strip().lower() != 'none'


##############################
# System -> Config utilities #
##############################
def set_hostname(browser, hostname):
    nav_to_module(browser, 'config')
    browser.find_by_id('id_hostname').fill(hostname)
    submit(browser, form_class='form-configuration')


def set_advanced_mode(browser, mode):
    nav_to_module(browser, 'config')
    advanced_mode = browser.find_by_id('id_advanced_mode')
    if mode:
        advanced_mode.check()
    else:
        advanced_mode.uncheck()

    submit(browser, form_class='form-configuration')


####################
# Backup utilities #
####################
def _click_button_and_confirm(browser, href, form_class):
    buttons = browser.links.find_by_href(href)
    if buttons:
        submit(browser, element=buttons.first)
        submit(browser, form_class=form_class,
               expected_url='/plinth/sys/backups/')


def _backup_delete_archive_by_name(browser, archive_name):
    nav_to_module(browser, 'backups')
    href = f'/plinth/sys/backups/root/delete/{archive_name}/'
    _click_button_and_confirm(browser, href, 'form-delete')


def backup_create(browser, app_name, archive_name=None):
    """Create a new backup for a given app."""
    install(browser, 'backups')
    if archive_name:
        _backup_delete_archive_by_name(browser, archive_name)

    buttons = browser.links.find_by_href('/plinth/sys/backups/create/')
    submit(browser, element=buttons.first)
    eventually(browser.find_by_css, args=['.select-all'])
    browser.find_by_css('.select-all').first.uncheck()
    if archive_name:
        browser.find_by_id('id_backups-name').fill(archive_name)

    # ensure the checkbox is scrolled into view
    browser.execute_script('window.scrollTo(0, 0)')
    browser.find_by_value(app_name).first.check()
    submit(browser, form_class='form-backups')


def backup_restore(browser, app_name, archive_name=None):
    """Restore a given app from a backup archive."""
    nav_to_module(browser, 'backups')
    href = f'/plinth/sys/backups/root/restore-archive/{archive_name}/'
    _click_button_and_confirm(browser, href, 'form-restore')


######################
# Networks utilities #
######################
def networks_set_firewall_zone(browser, zone):
    """"Set the network device firewall zone as internal or external."""
    nav_to_module(browser, 'networks')
    # First active Ethernet connection
    device = browser.find_by_xpath(
        '//span[contains(@class, "connection-type-label") and '
        'contains(., "Ethernet") ]/../..'
        '//span[contains(@class, "badge-success") '
        'and contains(@class, "connection-status-label")]/following::a').first
    network_id = device['href'].split('/')[-3]
    device.click()
    edit_url = "/plinth/sys/networks/{}/edit/".format(network_id)
    browser.links.find_by_href(edit_url).first.click()
    browser.select('zone', zone)
    submit(browser, form_class='form-connection-edit')


##################
# Bind utilities #
##################
def set_forwarders(browser, forwarders):
    """Set the forwarders list (space separated) in bind configuration."""
    nav_to_module(browser, 'bind')
    browser.fill('forwarders', forwarders)
    submit(browser, form_class='form-configuration')


def get_forwarders(browser):
    """Return the forwarders list (space separated) in bind configuration."""
    nav_to_module(browser, 'bind')
    return browser.find_by_name('forwarders').first.value


##############################
# Users and Groups utilities #
##############################


def create_user(browser, name, password=None, groups=[], email=None):
    """Create a user with password and user groups."""
    nav_to_module(browser, 'users')

    if password is None:
        password = get_password(name)

    with wait_for_page_update(browser):
        browser.links.find_by_href('/plinth/sys/users/create/').first.click()

    browser.find_by_id('id_username').fill(name)
    browser.find_by_id('id_password1').fill(password)
    browser.find_by_id('id_password2').fill(password)
    if email:
        browser.find_by_id('id_email').fill(email)

    for group in groups:
        browser.find_by_xpath(
            f'//label[contains(text(), "({group})")]/input').check()

    browser.find_by_id('id_confirm_password').fill(
        config['DEFAULT']['password'])

    submit(browser, form_class='form-create')


def delete_user(browser, name):
    """Delete a user."""
    nav_to_module(browser, 'users')
    delete_link = browser.links.find_by_href(
        f'/plinth/sys/users/{name}/delete/')

    with wait_for_page_update(browser):
        delete_link.first.click()
    submit(browser, form_class='form-delete')


def user_exists(browser, name):
    """Check if a user with a given name exists."""
    nav_to_module(browser, 'users')
    links = browser.links.find_by_href(f'/plinth/sys/users/{name}/edit/')
    return len(links) == 1


class BaseAppTests:
    """Base class for common functional tests.

    To run these tests on an app, the app should subclass this class with a
    class name starting with 'Test'. This allows the app tests to be
    automatically discovered while the base class tests to stay undiscovered.
    Tests will apply to the app without the need not override each test.
    """

    app_name = ''
    # TODO: Check the components of the app instead of configuring here.
    has_service = False
    has_web = True
    can_uninstall = True
    check_diagnostics = True
    diagnostics_delay = 0
    disable_after_tests = True

    def assert_app_running(self, session_browser):
        """Assert that the app is running."""
        if self.has_service:
            assert service_is_running(session_browser, self.app_name)

        if self.has_web:
            assert is_available(session_browser, self.app_name)

    def assert_app_not_running(self, session_browser):
        """Assert that the app is not running."""
        if self.has_service:
            assert service_is_not_running(session_browser, self.app_name)

        if self.has_web:
            assert not is_available(session_browser, self.app_name)

    def install_and_setup(self, session_browser):
        """Install the app and set it up if needed."""
        install(session_browser, self.app_name)

    @pytest.fixture(autouse=True)
    def fixture_background(self, session_browser):
        """Login, install, and enable the app."""
        login(session_browser)
        self.install_and_setup(session_browser)
        app_enable(session_browser, self.app_name)
        yield
        login(session_browser)
        if self.disable_after_tests:
            app_disable(session_browser, self.app_name)

    def test_enable_disable(self, session_browser):
        """Test enabling and disabling the app."""
        app_disable(session_browser, self.app_name)
        self.assert_app_not_running(session_browser)

        app_enable(session_browser, self.app_name)
        self.assert_app_running(session_browser)

    def test_run_diagnostics(self, session_browser):
        """Test that all app diagnostics are passing."""
        if not self.check_diagnostics:
            pytest.skip(f'Skipping diagnostics check for {self.app_name}.')

        time.sleep(self.diagnostics_delay)
        session_browser.find_by_id('id_extra_actions_button').click()
        submit(session_browser, form_class='form-diagnostics-button')

        warning_results = session_browser.find_by_css('.badge-warning')
        if warning_results:
            warnings.warn(
                f'Diagnostics warnings for {self.app_name}: {warning_results}')

        failure_results = session_browser.find_by_css('.badge-danger')
        assert not failure_results

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test that backup and restore operations work on the app."""
        backup_create(session_browser, self.app_name, 'test_' + self.app_name)
        if self.can_uninstall:
            uninstall(session_browser, self.app_name)

        backup_restore(session_browser, self.app_name, 'test_' + self.app_name)
        self.assert_app_running(session_browser)

    def test_uninstall(self, session_browser):
        """Test that app can be uninstalled and installed back again."""
        if not self.can_uninstall:
            pytest.skip(f'Skipping uninstall test for {self.app_name}.')

        uninstall(session_browser, self.app_name)
        assert not is_installed(session_browser, self.app_name)
        self.install_and_setup(session_browser)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for upgrades."""

import logging
import os
import subprocess

from aptsources import sourceslist
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop

import plinth
from plinth import app as app_module
from plinth import cfg, glib, kvstore, menu
from plinth.config import DropinConfigs
from plinth.daemon import RelatedDaemon
from plinth.modules.backups.components import BackupRestore
from plinth.package import Packages

from . import manifest, privileged

first_boot_steps = [
    {
        'id': 'backports_wizard',
        'url': 'upgrades:backports-firstboot',
        'order': 5,
    },
    {
        'id': 'initial_update',
        'url': 'upgrades:update-firstboot',
        'order': 6,
    },
]

_description = [
    _('Check for and apply the latest software and security updates.'),
    _('Updates are run at 06:00 everyday according to local time zone. Set '
      'your time zone in Date & Time app. Apps are restarted after update '
      'causing them to be unavailable briefly. If system reboot is deemed '
      'necessary, it is done automatically at 02:00 causing all apps to be '
      'unavailable briefly.')
]

BACKPORTS_REQUESTED_KEY = 'upgrades_backports_requested'

DIST_UPGRADE_ENABLED_KEY = 'upgrades_dist_upgrade_enabled'

logger = logging.getLogger(__name__)


class UpgradesApp(app_module.App):
    """FreedomBox app for software upgrades."""

    app_id = 'upgrades'

    _version = 18

    can_be_disabled = False

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True, name=_('Software Update'),
                               icon='fa-refresh', description=_description,
                               manual_page='Upgrades')
        self.add(info)

        menu_item = menu.Menu('menu-upgrades', info.name, None, info.icon,
                              'upgrades:index',
                              parent_url_name='system:system', order=50)
        self.add(menu_item)

        packages = Packages('packages-upgrades',
                            ['unattended-upgrades', 'needrestart'])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-upgrades', [
            '/etc/apt/apt.conf.d/20freedombox',
            '/etc/apt/apt.conf.d/20freedombox-allow-release-info-change',
            '/etc/apt/apt.conf.d/60unattended-upgrades',
            '/etc/needrestart/conf.d/freedombox.conf',
        ])
        self.add(dropin_configs)

        daemon = RelatedDaemon('related-daemon-upgrades',
                               'freedombox-dist-upgrade')
        self.add(daemon)

        backup_restore = BackupRestore('backup-restore-upgrades',
                                       **manifest.backup)
        self.add(backup_restore)

    def post_init(self):
        """Perform post initialization operations."""
        self._show_new_release_notification()

        # Check every day if backports becomes available, then configure it if
        # selected by user.
        glib.schedule(24 * 3600, setup_repositories)

        # Check every day if new stable release becomes available, then perform
        # dist-upgrade if updates are enabled
        glib.schedule(24 * 3600, check_dist_upgrade)

    def _show_new_release_notification(self):
        """When upgraded to new release, show a notification."""
        from plinth.notification import Notification
        try:
            note = Notification.get('upgrades-new-release')
            if note.data['version'] == plinth.__version__:
                # User already has notification for update to this version. It
                # may be dismissed or not yet dismissed
                return

            # User currently has a notification for an older version, update.
            dismiss = False
        except KeyError:
            # Don't show notification for the first version user runs, create
            # but don't show it.
            dismiss = True

        data = {
            'version': plinth.__version__,
            'app_name': 'translate:' + gettext_noop('Software Update'),
            'app_icon': 'fa-refresh'
        }
        title = gettext_noop('FreedomBox Updated')
        note = Notification.update_or_create(
            id='upgrades-new-release', app_id='upgrades', severity='info',
            title=title, body_template='upgrades-new-release.html', data=data,
            group='admin')
        note.dismiss(should_dismiss=dismiss)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)

        # Enable automatic upgrades but only on first install
        if not old_version and not cfg.develop:
            privileged.enable_auto()

        # Update apt preferences whenever on first install and on version
        # increment.
        privileged.setup()

        # When upgrading from a version without first boot wizard for
        # backports, assume backports have been requested.
        if old_version and old_version < 7:
            set_backports_requested(can_activate_backports())

        # Enable dist upgrade for new installs, and once when upgrading
        # from version without flag.
        if not old_version or old_version < 8:
            set_dist_upgrade_enabled(can_enable_dist_upgrade())

        # Try to setup apt repositories, if needed, if possible, on first
        # install and on version increment.
        setup_repositories(None)


def setup_repositories(_):
    """Setup apt repositories for backports."""
    if is_backports_requested():
        privileged.activate_backports(cfg.develop)


def check_dist_upgrade(_):
    """Check for upgrade to new stable release."""
    if is_dist_upgrade_enabled():
        try_start_dist_upgrade()


def try_start_dist_upgrade(test=False):
    """Try to start dist upgrade."""
    from plinth.notification import Notification

    result = privileged.start_dist_upgrade(test)
    dist_upgrade_started = result['dist_upgrade_started']
    reason = result['reason']
    if 'found-previous' in reason:
        logger.info(
            'Found previous dist-upgrade. If it was interrupted, it will '
            'be restarted.')
    elif 'already-' in reason:
        logger.info('Skip dist upgrade: System is already up-to-date.')
    elif 'codename-not-found' in reason:
        logger.warning('Skip dist upgrade: Codename not found in release '
                       'file.')
    elif 'upgrades-not-enabled' in reason:
        logger.info('Skip dist upgrade: Automatic updates are not enabled.')
    elif 'test-not-set' in reason:
        logger.info('Skip dist upgrade: --test is not set.')
    elif 'not-enough-free-space' in reason:
        logger.warning('Skip dist upgrade: Not enough free space in /.')
        title = gettext_noop('Could not start distribution update')
        message = gettext_noop(
            'There is not enough free space in the root partition to '
            'start the distribution update. Please ensure at least 5 GB '
            'is free. Distribution update will be retried after 24 hours,'
            ' if enabled.')
        Notification.update_or_create(id='upgrades-dist-upgrade-free-space',
                                      app_id='upgrades', severity='warning',
                                      title=title, message=message, actions=[{
                                          'type': 'dismiss'
                                      }], group='admin')
    elif 'started-dist-upgrade' in reason:
        logger.info('Started dist upgrade.')
        title = gettext_noop('Distribution update started')
        message = gettext_noop(
            'Started update to next stable release. This may take a long '
            'time to complete.')
        Notification.update_or_create(id='upgrades-dist-upgrade-started',
                                      app_id='upgrades', severity='info',
                                      title=title, message=message, actions=[{
                                          'type': 'dismiss'
                                      }], group='admin')
    else:
        logger.warning('Unhandled result of start-dist-upgrade: %s, %s',
                       dist_upgrade_started, reason)


def is_backports_requested():
    """Return whether user has chosen to activate backports."""
    return kvstore.get_default(BACKPORTS_REQUESTED_KEY, False)


def set_backports_requested(requested):
    """Set whether user has chosen to activate backports."""
    kvstore.set(BACKPORTS_REQUESTED_KEY, requested)
    logger.info('Backports requested - %s', requested)


def is_dist_upgrade_enabled():
    """Return whether user has enabled dist upgrade."""
    return kvstore.get_default(DIST_UPGRADE_ENABLED_KEY, False)


def set_dist_upgrade_enabled(enabled=True):
    """Set whether user has enabled dist upgrade."""
    kvstore.set(DIST_UPGRADE_ENABLED_KEY, enabled)
    logger.info('Distribution upgrade configured - %s', enabled)


def is_backports_enabled():
    """Return whether backports are enabled in the system configuration."""
    return os.path.exists(privileged.BACKPORTS_SOURCES_LIST)


def get_current_release():
    """Return current release and codename as a tuple."""
    output = subprocess.check_output(
        ['lsb_release', '--release', '--codename',
         '--short']).decode().strip()
    lines = output.split('\n')
    return lines[0], lines[1]


def is_backports_current():
    """Return whether backports are enabled for the current release."""
    if not is_backports_enabled():
        return False

    _, dist = get_current_release()
    dist += '-backports'
    sources = sourceslist.SourcesList()
    for source in sources:
        if source.dist == dist:
            return True

    return False


def can_activate_backports():
    """Return whether backports can be activated."""
    release, _ = get_current_release()
    if release == 'unstable' or (release == 'testing' and not cfg.develop):
        return False

    return True


def can_enable_dist_upgrade():
    """Return whether dist upgrade can be enabled."""
    release, _ = get_current_release()
    return release not in ['unstable', 'testing']


def can_test_dist_upgrade():
    """Return whether dist upgrade can be tested."""
    return can_enable_dist_upgrade() and cfg.develop


def test_dist_upgrade():
    """Test dist-upgrade from stable to testing."""
    if can_test_dist_upgrade():
        try_start_dist_upgrade(test=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""App manifest for Tor Proxy."""

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

_ORBOT_PACKAGE_ID = 'org.torproject.android'
_TOR_BROWSER_DOWNLOAD_URL = \
    'https://www.torproject.org/download/download-easy.html'

clients = [{
    'name':
        _('Tor Browser'),
    'platforms': [{
        'type': 'download',
        'os': 'windows',
        'url': _TOR_BROWSER_DOWNLOAD_URL,
    }, {
        'type': 'download',
        'os': 'gnu-linux',
        'url': _TOR_BROWSER_DOWNLOAD_URL,
    }, {
        'type': 'download',
        'os': 'macos',
        'url': _TOR_BROWSER_DOWNLOAD_URL,
    }]
}, {
    'name':
        _('Orbot: Proxy with Tor'),
    'platforms': [{
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', _ORBOT_PACKAGE_ID)
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'f-droid',
        'url': store_url('f-droid', _ORBOT_PACKAGE_ID)
    }]
}]

backup = {
    'config': {
        'directories': ['/etc/tor/instances/fbxproxy/'],
    },
    'secrets': {
        'directories': ['/var/lib/tor-instances/fbxproxy/']
    },
    'services': ['tor@fbxproxy']
}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for ttrss app.
"""

import pytest

from plinth.tests import functional

APP_ID = 'ttrss'

pytestmark = [pytest.mark.apps, pytest.mark.ttrss, pytest.mark.sso]


class TestTTRSSApp(functional.BaseAppTests):
    """Class to customize basic app tests for TTRSS."""

    app_name = 'ttrss'
    has_service = True
    has_web = True

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test backup and restore of app data."""
        functional.app_enable(session_browser, APP_ID)
        _subscribe(session_browser)
        functional.backup_create(session_browser, APP_ID, 'test_ttrss')

        functional.uninstall(session_browser, self.app_name)
        functional.backup_restore(session_browser, APP_ID, 'test_ttrss')

        assert functional.service_is_running(session_browser, APP_ID)
        assert _is_subscribed(session_browser)


def _ttrss_load_main_interface(browser):
    """Load the TT-RSS interface."""
    functional.visit(browser, '/tt-rss/')
    overlay = browser.find_by_id('overlay')
    functional.eventually(lambda: not overlay.visible)


def _expand_nodes(browser):
    """If interface has category nodes collapsed, expand them."""
    nodes = browser.find_by_css('span.dijitTreeExpandoClosed')
    for node in nodes:
        node.click()


def _is_feed_shown(browser, invert=False):
    """Return whether the test feed is present."""
    _expand_nodes(browser)
    return browser.is_text_present('Planet Debian') != invert


def _click_main_menu_item(browser, text):
    """Select an item from the main actions menu."""
    browser.find_by_css('.action-chooser').click()
    browser.find_by_text(text).click()


def _subscribe(browser):
    """Subscribe to a feed in TT-RSS."""

    def _already_subscribed_message():
        return browser.is_text_present(
            'You are already subscribed to this feed.')

    _ttrss_load_main_interface(browser)

    _click_main_menu_item(browser, 'Subscribe to feed...')
    browser.find_by_id('feedDlg_feedUrl').fill(
        'https://planet.debian.org/atom.xml')
    browser.find_by_text('Subscribe').click()
    add_dialog = browser.find_by_css('#feedAddDlg')
    functional.eventually(
        lambda: not add_dialog.visible or _already_subscribed_message())
    if _already_subscribed_message():
        browser.find_by_text('Cancel').click()
        functional.eventually(lambda: not add_dialog.visible)


def _unsubscribe(browser):
    """Unsubscribe from a feed in TT-RSS."""
    _ttrss_load_main_interface(browser)
    _expand_nodes(browser)

    browser.find_by_text('Planet Debian').click()
    _click_main_menu_item(browser, 'Unsubscribe')

    prompt = browser.get_alert()
    prompt.accept()

    # Reload as sometimes the feed does not disappear immediately
    _ttrss_load_main_interface(browser)

    assert functional.eventually(_is_feed_shown, [browser, True])


def _is_subscribed(browser):
    """Return whether subscribed to a feed in TT-RSS."""
    _ttrss_load_main_interface(browser)
    return _is_feed_shown(browser)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""URLs for the Nextcloud module."""

from django.urls import re_path

from .views import NextcloudAppView

urlpatterns = [
    re_path(r'^apps/nextcloud/$', NextcloudAppView.as_view(), name='index')
]
